# Copyright(C) 2010-2012 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from copy import copy

from woob.core import CallErrors
from woob.applications.msg import AppMsg
from woob.capabilities.dating import CapDating, OptimizationNotFound
from woob.tools.application.formatters.iformatter import PrettyFormatter


__all__ = ['AppDating']


class EventListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('date', 'type')

    def get_title(self, event):
        s = '(%s) %s' % (event.date, event.type)
        if hasattr(event, 'contact') and event.contact:
            s += ' — %s (%s)' % (event.contact.name, event.contact.id)

        return s

    def get_description(self, event):
        if hasattr(event, 'message'):
            return event.message


class AppDating(AppMsg):
    APPNAME = 'dating'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to interact with various dating websites " \
                  "and to optimize seduction algorithmically."
    SHORT_DESCRIPTION = "interact with dating websites"
    STORAGE_FILENAME = 'dating.storage'
    STORAGE = {'optims': {}}
    CAPS = CapDating
    EXTRA_FORMATTERS = copy(AppMsg.EXTRA_FORMATTERS)
    EXTRA_FORMATTERS['events'] = EventListFormatter
    COMMANDS_FORMATTERS = copy(AppMsg.COMMANDS_FORMATTERS)
    COMMANDS_FORMATTERS['optim'] = 'table'
    COMMANDS_FORMATTERS['events'] = 'events'

    def load_default_backends(self):
        self.load_backends(CapDating, storage=self.create_storage(self.STORAGE_FILENAME))

    def main(self, argv):
        self.load_config()

        try:
            self._do_and_retry_wait('init_optimizations')
        except CallErrors as e:
            self.bcall_errors_handler(e)

        optimizations = self.storage.get('optims')
        for optim, backends in optimizations.items():
            self.optims('start', backends, optim, store=False)

        return super().main(argv)

    def do_query(self, id):
        """
        query ID

        Send a query to someone.
        """
        _id, backend_name = self.parse_id(id, unique_backend=True)

        for query in self.do('send_query', _id, backends=backend_name):
            print('%s' % query.message)

    def edit_optims(self, backend_names, optims_names, stop=False):
        if optims_names is None:
            print('Error: missing parameters.', file=self.stderr)
            return 2

        for optim_name in optims_names.split():
            backends_optims = {}
            for optim in self.do('get_optimization', optim_name, backends=backend_names):
                if optim:
                    backends_optims[optim.backend] = optim
            for backend_name, optim in backends_optims.items():
                if len(optim.CONFIG) == 0:
                    print('%s.%s does not require configuration.' % (backend_name, optim_name))
                    continue

                was_running = optim.is_running()
                if stop and was_running:
                    print('Stopping %s: %s' % (optim_name, backend_name))
                    optim.stop()
                params = optim.get_config()
                if params is None:
                    params = {}
                print('Configuration of %s.%s' % (backend_name, optim_name))
                print('-----------------%s-%s' % ('-' * len(backend_name), '-' * len(optim_name)))
                for key, value in optim.CONFIG.items():
                    params[key] = self.ask(value, default=params[key] if (key in params) else value.default)

                optim.set_config(params)
                if stop and was_running:
                    print('Starting %s: %s' % (optim_name, backend_name))
                    optim.start()

    def optims(self, function, backend_names, optims, store=True):
        if optims is None:
            print('Error: missing parameters.', file=self.stderr)
            return 2

        for optim_name in optims.split():
            try:
                if store:
                    storage_optim = set(self.storage.get('optims', optim_name, default=[]))
                self.stdout.write('%sing %s:' % (function.capitalize(), optim_name))
                for optim in self.do('get_optimization', optim_name, backends=backend_names):
                    if optim:
                        # It's useless to start a started optim, or to stop a stopped one.
                        if (function == 'start' and optim.is_running()) or \
                           (function == 'stop' and not optim.is_running()):
                            continue

                        # Optim is not configured and would be, ask user to do it.
                        if function == 'start' and len(optim.CONFIG) > 0 and optim.get_config() is None:
                            self.edit_optims(optim.backend, optim_name)

                        ret = getattr(optim, function)()
                        self.stdout.write(' ' + optim.backend)
                        if not ret:
                            self.stdout.write('(failed)')
                        self.stdout.flush()
                        if store:
                            if function == 'start' and ret:
                                storage_optim.add(optim.backend)
                            elif function == 'stop':
                                try:
                                    storage_optim.remove(optim.backend)
                                except KeyError:
                                    pass
                self.stdout.write('.\n')
            except CallErrors as errors:
                for backend, error, backtrace in errors:
                    if isinstance(error, OptimizationNotFound):
                        self.logger.error('Error(%s): Optimization "%s" not found' % (backend.name, optim_name))
                    else:
                        self.bcall_error_handler(backend, error, backtrace)
            if store:
                if len(storage_optim) > 0:
                    self.storage.set('optims', optim_name, list(storage_optim))
                else:
                    self.storage.delete('optims', optim_name)
        if store:
            self.storage.save()

    def complete_optim(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return ['list', 'start', 'stop', 'edit']
        elif len(args) == 3:
            return [backend.name for backend in self.enabled_backends]
        elif len(args) >= 4:
            if args[2] == '*':
                backend = None
            else:
                backend = args[2]
            optims = set()
            for optim in self.do('iter_optimizations', backends=backend):
                optims.add(optim.id)
            return sorted(optims - set(args[3:]))

    def do_optim(self, line):
        """
        optim [list | start | edit | stop] BACKEND [OPTIM1 [OPTIM2 ...]]

        All dating backends offer optimization services. This command can be
        manage them.
        Use * us BACKEND value to apply command to all backends.

        Commands:
        * list       list all available optimizations of a backend
        * start      start optimization services on a backend
        * edit       configure an optimization service for a backend
        * stop       stop optimization services on a backend
        """
        cmd, backend_name, optims_names = self.parse_command_args(line, 3)

        if backend_name == '*':
            backend_name = None
        elif backend_name is not None and backend_name not in [b.name for b in self.enabled_backends]:
            print('Error: No such backend "%s"' % backend_name, file=self.stderr)
            return 1

        if cmd == 'start':
            return self.optims('start', backend_name, optims_names)
        if cmd == 'stop':
            return self.optims('stop', backend_name, optims_names)
        if cmd == 'edit':
            self.edit_optims(backend_name, optims_names, stop=True)
            return
        if cmd == 'list' or cmd is None:
            if optims_names is not None:
                optims_names = optims_names.split()

            optims = {}
            backends = set()
            for optim in self.do('iter_optimizations', backends=backend_name):
                if optims_names is not None and optim.id not in optims_names:
                    continue
                if optim.is_running():
                    status = 'RUNNING'
                else:
                    status = '-------'
                if optim.id not in optims:
                    optims[optim.id] = {optim.backend: status}
                else:
                    optims[optim.id][optim.backend] = status
                backends.add(optim.backend)

            backends = sorted(backends)
            for name, backends_status in optims.items():
                line = [('name', name)]
                for b in backends:
                    try:
                        status = backends_status[b]
                    except KeyError:
                        status = ''
                    line.append((b, status))
                self.format(tuple(line))
            return
        print("No such command '%s'" % cmd, file=self.stderr)
        return 1

    def do_events(self, line):
        """
        events

        Display dating events.
        """
        self.change_path(['events'])
        self.start_format()
        for event in self.do('iter_events'):
            self.cached_format(event)



# Copyright(C) 2010-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, Field, StringField, FloatField, \
                  IntField, UserError
from .date import DateField

__all__ = ['MagnetOnly', 'Torrent', 'CapTorrent']


class MagnetOnly(UserError):
    """
    Raised when trying to get URL to torrent but only magnet is available.
    """

    def __init__(self, magnet):
        self.magnet = magnet
        super(MagnetOnly, self).__init__('Only magnet URL is available')


class Torrent(BaseObject):
    """
    Torrent object.
    """
    name =      StringField('Name of torrent')
    size =      FloatField('Size of torrent')
    date =      DateField('Date when torrent has been published')
    magnet =    StringField('URI of magnet')
    seeders =   IntField('Number of seeders')
    leechers =  IntField('Number of leechers')
    files =     Field('Files in torrent', list)
    description =   StringField('Description of torrent')
    filename =      StringField('Name of .torrent file')


class CapTorrent(Capability):
    """
    Torrent trackers.
    """

    def iter_torrents(self, pattern):
        """
        Search torrents and iterate on results.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`Torrent`]
        """
        raise NotImplementedError()

    def get_torrent(self, _id):
        """
        Get a torrent object from an ID.

        :param _id: ID of torrent
        :type _id: str
        :rtype: :class:`Torrent`
        """
        raise NotImplementedError()

    def get_torrent_file(self, _id):
        """
        Get the content of the .torrent file.

        :param _id: ID of torrent
        :type _id: str
        :rtype: bytes
        """
        raise NotImplementedError()



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import os
import tempfile

import woob.tools.date
import yaml

from .iconfig import ConfigError, IConfig
from .util import LOGGER, replace

try:
    from yaml import CSafeLoader as SafeLoader
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import SafeLoader
    from yaml import Dumper


__all__ = ['YamlConfig']


class WoobDumper(Dumper):
    pass


WeboobDumper = WoobDumper


class WoobNoAliasDumper(WoobDumper):
    def ignore_aliases(self, data):
        return True


WeboobNoAliasDumper = WoobNoAliasDumper


WoobDumper.add_representer(woob.tools.date.date,
                           WoobDumper.represent_date)

WoobDumper.add_representer(woob.tools.date.datetime,
                           WoobDumper.represent_datetime)


class YamlConfig(IConfig):
    DUMPER = WoobDumper
    LOADER = SafeLoader

    def __init__(self, path):
        self.path = path
        self.values = {}

    def load(self, default={}):
        self.values = default.copy()

        LOGGER.debug('Loading configuration file: %s.' % self.path)
        try:
            with open(self.path, 'r') as f:
                self.values = yaml.load(f, Loader=self.LOADER)  # nosec: bandit can't detect SafeLoader…
            LOGGER.debug('Configuration file loaded: %s.' % self.path)
        except IOError:
            self.save()
            LOGGER.debug('Configuration file created with default values: %s.' % self.path)

        if self.values is None:
            self.values = {}

    def save(self):
        # write in a temporary file to avoid corruption problems
        f = tempfile.NamedTemporaryFile(mode='w', dir=os.path.dirname(self.path), delete=False, encoding='utf-8')
        with f:
            yaml.dump(self.values, f, Dumper=self.DUMPER, default_flow_style=False)
        replace(f.name, self.path)
        LOGGER.debug('Configuration file saved: %s.' % self.path)

    def get(self, *args, **kwargs):
        v = self.values
        for a in args[:-1]:
            try:
                v = v[a]
            except KeyError:
                if 'default' in kwargs:
                    return kwargs['default']
                else:
                    raise ConfigError()
            except TypeError:
                raise ConfigError()

        try:
            v = v[args[-1]]
        except KeyError:
            v = kwargs.get('default')

        return v

    def set(self, *args):
        v = self.values
        for a in args[:-2]:
            try:
                v = v[a]
            except KeyError:
                v[a] = {}
                v = v[a]
            except TypeError:
                raise ConfigError()

        v[args[-2]] = args[-1]

    def delete(self, *args):
        v = self.values
        for a in args[:-1]:
            try:
                v = v[a]
            except KeyError:
                return
            except TypeError:
                raise ConfigError()

        v.pop(args[-1], None)



# Copyright(C) 2012-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from collections import OrderedDict
import warnings

from urllib3.util.request import ACCEPT_ENCODING
ENCODINGS = [e.strip() for e in ACCEPT_ENCODING.split(',')]

from woob import __version__


class Profile:
    """
    A profile represents the way Browser should act.
    Usually it is to mimic a real browser.
    """

    def setup_session(self, session):
        """
        Change default headers, set up hooks, etc.

        Warning: Do not enable lzma, bzip or bzip2, sdch encodings
        as python-requests does not support it yet.
        Supported as of 2.2: gzip, deflate, compress.
        In doubt, do not change the default Accept-Encoding header
        of python-requests.
        """
        raise NotImplementedError()


class Weboob(Profile):
    """
    .. deprecated:: 3.4
       Don't use this profile, use :class:`Woob` instead.
    """

    def __init__(self, version=None):
        warnings.warn(
            'This profile is deprecated and will be removed in woob 4.0. Use the Woob profile instead',
            DeprecationWarning
        )
        self.version = version or __version__

    def setup_session(self, session):
        session.headers['User-Agent'] = 'weboob/%s' % self.version


class Woob(Profile):
    """
    It's us!
    Recommended for Woob-friendly websites only.
    """

    def __init__(self, version=None):
        self.version = version or __version__

    def setup_session(self, session):
        session.headers['User-Agent'] = 'woob/%s' % self.version


class Firefox(Profile):
    """
    Try to mimic a specific version of Firefox.
    Ideally, it should follow the current ESR Firefox:
    https://www.mozilla.org/en-US/firefox/organizations/all.html
    Do not change the Firefox version without checking the Gecko one!
    """

    def setup_session(self, session):
        """
        Set up headers for a standard Firefox request
        (except for DNT which isn't on by default but is a good idea).

        The goal is to be unidentifiable.
        """
        # Replace all base requests headers
        # https://developer.mozilla.org/en/Gecko_user_agent_string_reference
        # https://bugzilla.mozilla.org/show_bug.cgi?id=572650
        session.headers = OrderedDict([
            ('Accept-Language', 'en-US,en;q=0.5'),
            ('Accept-Encoding', 'gzip, deflate'),
            ('Accept', 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'),
            ('User-Agent', 'Mozilla/5.0 (X11; Linux x86_64; rv:102.0) Gecko/20100101 Firefox/102.0'),
            ('Upgrade-Insecure-Requests', '1'),
            ('DNT', '1'),
        ])

        if 'br' in ENCODINGS:
            session.headers['Accept-Encoding'] += ', br'


class GoogleBot(Profile):
    """
    Try to mimic Googlebot.
    Keep in mind there are ways to authenticate real Googlebot IPs.

    You will most likely want to set ALLOW_REFERRER to False.
    """

    def setup_session(self, session):
        """
        Set up headers for a standard Firefox request
        (except for DNT which isn't on by default but is a good idea).

        The goal is to be unidentifiable.
        """
        # Replace all base requests headers
        # http://googlewebmastercentral.blogspot.com/2008/03/first-date-with-googlebot-headers-and.html
        # Cached versions of:
        #  http://request.urih.com/
        #  http://xhaus.com/headers
        session.headers = {
            'Accept-Encoding': 'gzip,deflate',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'From': 'googlebot(at)googlebot.com',
            'User-Agent': 'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)'}


class Wget(Profile):
    """
    Common alternative user agent.
    Some websites will give you a version with less JavaScript.
    Some others could ban you (after all, wget is not a real browser).
    """

    def __init__(self, version='1.11.4'):
        self.version = version

    def setup_session(self, session):
        # Don't remove base headers, if websites want to block fake browsers,
        # they will probably block any wget user agent anyway.
        session.headers.update({
            'Accept': '*/*',
            'User-Agent': 'Wget/%s' % self.version})


class Android(Profile):
    """
    An android profile for mobile websites
    """

    def setup_session(self, session):
        """
        Set up user agent.
        """
        session.headers.update({
            'User-Agent': 'Mozilla/5.0 (Linux; U; Android 4.0.3; fr-fr; LG-L160L Build/IML74K) AppleWebkit/534.30 (KHTML, like Gecko) Version/4.0 Mobile Safari/534.30',
            'Upgrade-Insecure-Requests': '1',
        })


class IPhone(Profile):
    """
    An iphone profile for mobile websites and some API websites
    """

    def __init__(self, application):
        self.application = application

    def setup_session(self, session):
        session.headers["Accept-Language"] = "en;q=1, fr;q=0.9, de;q=0.8, ja;q=0.7, nl;q=0.6, it;q=0.5"
        session.headers["Accept"] = "*/*"
        session.headers["User-Agent"] = "%s (iPhone; iOS 7.1; Scale/2.00)" % self.application
        session.headers["Accept-Encoding"] = "gzip, deflate"
        session.headers["Upgrade-Insecure-Requests"] = '1'



# Copyright(C) 2012-2022 woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import base64
import io
import os
from datetime import datetime
from threading import Lock
from urllib.parse import urlparse, parse_qsl

from woob.tools.json import json
from woob.tools.log import getLogger
from woob import __version__ as woob_version

__all__ = ['HARManager']


class HARManager:
    def __init__(self, responses_dirname, logger):
        self.har_path = os.path.join(responses_dirname, 'bundle.har')
        self.responses_lock = Lock()
        self.logger = getLogger('har', logger)

        self.bundle = None

    def _build_har_bundle(self, started_datetime):
        self.bundle = {
            'log': {
                'version': '1.2',
                'creator': {
                    'name': 'woob',
                    'version': woob_version,
                },
                'browser': {
                    'name': 'woob',
                    'version': woob_version,
                },
                # there are no pages, but we need that to please firefox
                'pages': [{
                    'id': 'fake_page',
                    'pageTimings': {},
                    # and chromium wants some of it too
                    'startedDateTime': started_datetime,
                }],
                # don't put additional data after this list, to have a fixed-size suffix after it
                # so we can add more entries without rewriting the whole file.
                'entries': [],
            },
        }

    @staticmethod
    def _build_har_request(request, http_version):
        request_entry = {
            'method': request.method,
            'url': request.url,
            'httpVersion': http_version,
            'headers': [
                {
                    'name': k,
                    'value': v,
                }
                for k, v in request.headers.items()
            ],
            'queryString': [
                {
                    'name': key,
                    'value': value,
                }
                for key, value in parse_qsl(
                    urlparse(request.url).query,
                    keep_blank_values=True,
                )
            ],
            'cookies': [
                {
                    'name': k,
                    'value': v,
                }
                for k, v in request._cookies.items()
            ],
            # for chromium
            'bodySize': -1,
            'headersSize': -1,
        }

        if request.body is not None:
            request_entry['postData'] = {
                'mimeType': request.headers.get('Content-Type', ''),
                'params': [],
            }
            if isinstance(request.body, str):
                request_entry['postData']['text'] = request.body
            else:
                # HAR format has no proper way to encode posted binary data!
                request_entry['postData']['text'] = request.body.decode('latin-1')
                # add a non-standard key to indicate how should "text" be decoded.
                request_entry['postData']['x-binary'] = True

            if request.headers.get('Content-Type') == 'application/x-www-form-urlencoded':
                request_entry['postData']['params'] = [
                    {
                        "name": key,
                        "value": value,
                    } for key, value in parse_qsl(request.body)
                ]

        return request_entry

    @staticmethod
    def _build_har_response(response):
        response_entry = {
            'status': response.status_code,
            'statusText': response.reason,
            'httpVersion': 'HTTP/%.1f' % (response.raw.version / 10.),
            'headers': [
                {
                    'name': k,
                    'value': v,
                }
                for k, v in response.headers.items()
            ],
            'content': {
                'mimeType': response.headers.get('Content-Type', ''),
                'size': len(response.content),
                # systematically use base64 to avoid more content alteration
                # than there already is...
                'encoding': "base64",
                'text': base64.b64encode(response.content).decode('ascii'),
            },
            'cookies': [
                {
                    'name': k,
                    'value': v,
                }
                for k, v in response.cookies.items()
            ],
            'redirectURL': response.headers.get('location', ''),
            # for chromium
            'bodySize': -1,
            'headersSize': -1,
        }
        return response_entry

    @staticmethod
    def _build_empty_har_response(*args):
        # called when we get a timeout
        return {
            'status': 0,
            'statusText': '',
            'httpVersion': '',
            'headers': [],
            'content': {},
            'cookies': [],
            'redirectURL': '',
            # for chromium
            'bodySize': -1,
            'headersSize': -1,
        }

    def _build_har_entry(self, slug, request, response=None, time=''):
        # check if response is not None and not if response
        # because a response with a status_code >= 400 is falsy
        if response is not None:
            started_datetime = (datetime.now() - response.elapsed).isoformat()
            time = int(response.elapsed.total_seconds() * 1000)
            http_version = 'HTTP/%.1f' % (response.raw.version / 10.)
            build_response = self._build_har_response
        else:
            started_datetime = datetime.now().isoformat()
            build_response = self._build_empty_har_response
            http_version = ''

        if not self.bundle:
            self._build_har_bundle(started_datetime)

        har_entry = {
            '$anchor': slug,
            'startedDateTime': started_datetime,
            'pageref': 'fake_page',
            'time': time,
            'request': self._build_har_request(request, http_version),
            'response': build_response(response),
            'timings': {  # please chromium
                'send': -1,
                'wait': -1,
                'receive': -1,
            },
            'cache': {},
        }
        return har_entry

    def _save_har_entry(self, har_entry):
        self.bundle['log']['entries'].append(har_entry)

        if not os.path.isfile(self.har_path):
            with open(self.har_path, 'w') as fd:
                json.dump(self.bundle, fd, separators=(',', ':'))
        else:
            # hack to avoid rewriting the whole file: entries are last in the JSON file
            # we need to seek at the right place and write the new entry.
            # this will unfortunately overwrite closings.
            suffix = "]}}"
            with open(self.har_path, 'r+') as fd:
                # can't seek with a negative value...
                fd.seek(0, io.SEEK_END)
                after_entry_pos = fd.tell() - len(suffix)
                fd.seek(after_entry_pos)

                if fd.read(len(suffix)) != suffix:
                    self.logger.warning('HAR file does not end with the expected pattern')
                else:
                    fd.seek(after_entry_pos)
                    fd.write(',')  # there should have been at least one entry
                    json.dump(har_entry, fd, separators=(',', ':'))
                    fd.write(suffix)

    def save_response(self, slug, response):
        request = response.request
        har_entry = self._build_har_entry(slug, request, response=response)

        with self.responses_lock:
            self._save_har_entry(har_entry)

    def save_request_only(self, slug, request, time):
        har_entry = self._build_har_entry(slug, request, time=time)
        with self.responses_lock:
            self._save_har_entry(har_entry)



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .iformatter import IFormatter


__all__ = ['SimpleFormatter']


class SimpleFormatter(IFormatter):
    def __init__(self, field_separator='\t', key_value_separator='='):
        super(SimpleFormatter, self).__init__()
        self.field_separator = field_separator
        self.key_value_separator = key_value_separator

    def format_dict(self, item):
        return self.field_separator.join('%s%s' % (
            ('%s%s' % (k, self.key_value_separator) if self.display_keys else ''), v)
            for k, v in item.items())



# Copyright(C) 2011  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from datetime import timedelta
from email import message_from_string, message_from_file
from email.header import decode_header
from email.mime.text import MIMEText
from smtplib import SMTP
import os
import re
import unicodedata

from woob.capabilities.base import empty, BaseObject
from woob.capabilities.bugtracker import CapBugTracker, Query, Update, Project, Issue, IssueError
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.tools.html import html2text
from woob.tools.date import parse_french_date


__all__ = ['AppBugTracker']


class IssueFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'project', 'title', 'body', 'author')

    def format_attr(self, obj, attr):
        if not hasattr(obj, attr) or empty(getattr(obj, attr)):
            return ''

        value = getattr(obj, attr)
        if isinstance(value, BaseObject):
            value = value.name

        return self.format_key(attr.capitalize(), value)

    def format_key(self, key, value):
        return '%s %s\n' % (self.colored('%s:' % key, 'green'),
                            value)

    def format_obj(self, obj, alias):
        result = '%s %s %s %s %s\n' % (self.colored(obj.project.name, 'blue', 'bold'),
                                        self.colored('—', 'cyan', 'bold'),
                                        self.colored(obj.fullid, 'red', 'bold'),
                                        self.colored('—', 'cyan', 'bold'),
                                        self.colored(obj.title, 'yellow', 'bold'))
        result += '\n%s\n\n' % obj.body
        result += self.format_key('Author', '%s (%s)' % (obj.author.name, obj.creation))
        result += self.format_attr(obj, 'status')
        result += self.format_attr(obj, 'priority')
        result += self.format_attr(obj, 'version')
        result += self.format_attr(obj, 'tracker')
        result += self.format_attr(obj, 'category')
        result += self.format_attr(obj, 'assignee')
        if hasattr(obj, 'fields') and not empty(obj.fields):
            for key, value in obj.fields.items():
                result += self.format_key(key.capitalize(), value)
        if hasattr(obj, 'attachments') and obj.attachments:
            result += '\n%s\n' % self.colored('Attachments:', 'green')
            for a in obj.attachments:
                result += '* %s%s%s <%s>\n' % (self.BOLD, a.filename, self.NC, a.url)
        if hasattr(obj, 'history') and obj.history:
            result += '\n%s\n' % self.colored('History:', 'green')
            for u in obj.history:
                result += '%s %s %s %s\n' % (self.colored('*', 'red', 'bold'),
                                             self.colored(u.date, 'yellow', 'bold'),
                                             self.colored('—', 'cyan', 'bold'),
                                             self.colored(u.author.name, 'blue', 'bold'))
                for change in u.changes:
                    result += '  - %s %s %s %s\n' % (self.colored(change.field, 'green'),
                                                     change.last,
                                                     self.colored('->', 'magenta'), change.new)
                if u.message:
                    result += '    %s\n' % html2text(u.message).strip().replace('\n', '\n    ')
        return result


class IssuesListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'project', 'status', 'title', 'category')

    def get_title(self, obj):
        return '%s - [%s] %s' % (obj.project.name, obj.status.name, obj.title)

    def get_description(self, obj):
        return obj.category


class AppBugTracker(ReplApplication):
    APPNAME = 'bugtracker'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2011-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to create, edit, view bug tracking issues."
    SHORT_DESCRIPTION = "manage bug tracking issues"
    CAPS = CapBugTracker
    EXTRA_FORMATTERS = {'issue_info': IssueFormatter,
                        'issues_list': IssuesListFormatter,
                       }
    COMMANDS_FORMATTERS = {'get':     'issue_info',
                           'post':    'issue_info',
                           'edit':    'issue_info',
                           'search':  'issues_list',
                           'ls':      'issues_list',
                          }
    COLLECTION_OBJECTS = (Project, Issue, )

    def add_application_options(self, group):
        group.add_option('--author')
        group.add_option('--title')
        group.add_option('--assignee')
        group.add_option('--target-version', dest='version')
        group.add_option('--tracker')
        group.add_option('--category')
        group.add_option('--status')
        group.add_option('--priority')
        group.add_option('--start')
        group.add_option('--due')

    @defaultcount(10)
    def do_search(self, line):
        """
        search PROJECT

        List issues for a project.

        You can use these filters from command line:
           --author AUTHOR
           --title TITLE_PATTERN
           --assignee ASSIGNEE
           --target-version VERSION
           --category CATEGORY
           --status STATUS
        """
        query = Query()

        path = self.working_path.get()
        backends = []
        if line.strip():
            query.project, backends = self.parse_id(line, unique_backend=True)
        elif len(path) > 0:
            query.project = path[0]
        else:
            print('Please enter a project name', file=self.stderr)
            return 1

        query.author = self.options.author
        query.title = self.options.title
        query.assignee = self.options.assignee
        query.version = self.options.version
        query.category = self.options.category
        query.status = self.options.status

        self.change_path([query.project, 'search'])
        for issue in self.do('iter_issues', query, backends=backends):
            self.add_object(issue)
            self.format(issue)

    def complete_get(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_get(self, line):
        """
        get ISSUE

        Get an issue and display it.
        """
        if not line:
            print('This command takes an argument: %s' % self.get_command_help('get', short=True), file=self.stderr)
            return 2

        issue = self.get_object(line, 'get_issue')
        if not issue:
            print('Issue not found: %s' % line, file=self.stderr)
            return 3
        self.format(issue)

    def complete_comment(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_comment(self, line):
        """
        comment ISSUE [TEXT]

        Comment an issue. If no text is given, enter it in standard input.
        """
        id, text = self.parse_command_args(line, 2, 1)
        if text is None:
            text = self.acquire_input()

        id, backend_name = self.parse_id(id, unique_backend=True)
        update = Update(0)
        update.message = text

        self.do('update_issue', id, update, backends=backend_name).wait()

    def do_logtime(self, line):
        """
        logtime ISSUE HOURS [TEXT]

        Log spent time on an issue.
        """
        id, hours, text = self.parse_command_args(line, 3, 2)
        if text is None:
            text = self.acquire_input()

        try:
            hours = float(hours)
        except ValueError:
            print('Error: HOURS parameter may be a float', file=self.stderr)
            return 1

        id, backend_name = self.parse_id(id, unique_backend=True)
        update = Update(0)
        update.message = text
        update.hours = timedelta(hours=hours)

        self.do('update_issue', id, update, backends=backend_name).wait()

    def complete_remove(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_remove(self, line):
        """
        remove ISSUE

        Remove an issue.
        """
        id, backend_name = self.parse_id(line, unique_backend=True)
        self.do('remove_issue', id, backends=backend_name).wait()

    ISSUE_FIELDS = (('title',    (None,       False)),
                    ('assignee', ('members',  True)),
                    ('version',  ('versions', True)),
                    ('tracker',  (None,       False)),#XXX
                    ('category', ('categories', False)),
                    ('status',   ('statuses', True)),
                    ('priority', (None,       False)),#XXX
                    ('start',    (None,       False)),
                    ('due',      (None,       False)),
                   )

    def get_list_item(self, objects_list, name):
        if name is None:
            return None

        for obj in objects_list:
            if obj.name.lower() == name.lower():
                return obj

        if not name:
            return None

        raise ValueError('"%s" is not found' % name)

    def sanitize_key(self, key):
        if isinstance(key, str):
            key = str(key, "utf8")
        key = unicodedata.normalize('NFKD', key).encode("ascii", "ignore")
        return key.replace(' ', '-').capitalize()

    def issue2text(self, issue, backend=None):
        if backend is not None and 'username' in backend.config:
            sender = backend.config['username'].get()
        else:
            sender = os.environ.get('USERNAME', 'bugtracker')
        output = 'From: %s\n' % sender
        for key, (list_name, is_list_object) in self.ISSUE_FIELDS:
            value = None
            if not self.interactive:
                value = getattr(self.options, key)
            if not value:
                value = getattr(issue, key)
            if not value:
                value = ''
            elif hasattr(value, 'name'):
                value = value.name

            if list_name is not None:
                objects_list = getattr(issue.project, list_name)
                if len(objects_list) == 0:
                    continue

            output += '%s: %s\n' % (self.sanitize_key(key), value)
            if list_name is not None:
                availables = ', '.join('<%s>' % (o if isinstance(o, str) else o.name)
                                       for o in objects_list)
                output += 'X-Available-%s: %s\n' % (self.sanitize_key(key), availables)

        for key, value in issue.fields.items():
            output += '%s: %s\n' % (self.sanitize_key(key), value or '')
            # TODO: Add X-Available-* for lists

        output += '\n%s' % (issue.body or 'Please write your bug report here.')
        return output

    def text2issue(self, issue, m):
        # XXX HACK to support real incoming emails
        if 'Subject' in m:
            m['Title'] = m['Subject']

        for key, (list_name, is_list_object) in self.ISSUE_FIELDS:
            value = m.get(key)
            if value is None:
                continue

            new_value = ''
            for part in decode_header(value):
                if part[1]:
                    new_value += str(part[0], part[1])
                else:
                    new_value += part[0].decode('utf-8')
            value = new_value

            if is_list_object:
                objects_list = getattr(issue.project, list_name)
                value = self.get_list_item(objects_list, value)

            # FIXME: autodetect
            if key in ['start', 'due']:
                if len(value) > 0:
                    #value = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
                    value = parse_french_date(value)
                else:
                    value = None

            setattr(issue, key, value)

        for key in issue.fields.keys():
            value = m.get(self.sanitize_key(key))
            if value is not None:
                issue.fields[key] = value.decode('utf-8')

        content = ''
        for part in m.walk():
            if part.get_content_type() == 'text/plain':
                s = part.get_payload(decode=True)
                charsets = part.get_charsets() + m.get_charsets()
                for charset in charsets:
                    try:
                        if charset is not None:
                            content += str(s, charset)
                        else:
                            content += str(s, encoding='utf-8')
                    except UnicodeError as e:
                        self.logger.warning('Unicode error: %s' % e)
                        continue
                    except Exception as e:
                        self.logger.exception(e)
                        continue
                    else:
                        break

        issue.body = content

        m = re.search('([^< ]+@[^ >]+)', m['From'] or '')
        if m:
            return m.group(1)

    def edit_issue(self, issue, edit=True):
        backend = self.woob.get_backend(issue.backend)
        content = self.issue2text(issue, backend)
        while True:
            if self.stdin.isatty():
                content = self.acquire_input(content, {'vim': "-c 'set ft=mail'"})
                m = message_from_string(content.encode('utf-8'))
            else:
                m = message_from_file(self.stdin)

            try:
                email_to = self.text2issue(issue, m)
            except ValueError as e:
                if not self.stdin.isatty():
                    raise
                input("%s -- Press Enter to continue..." % e)
                continue

            try:
                issue = backend.post_issue(issue)
                print('Issue %s %s' % (self.formatter.colored(issue.fullid, 'red', 'bold'),
                                       'updated' if edit else 'created'))
                if edit:
                    self.format(issue)
                elif email_to:
                    self.send_notification(email_to, issue)
                return 0
            except IssueError as e:
                if not self.stdin.isatty():
                    raise
                input("%s -- Press Enter to continue..." % e)

    def send_notification(self, email_to, issue):
        text = """Hi,

You have successfuly created this ticket on the Woob tracker:

%s

You can follow your bug report on this page:

https://symlink.me/issues/%s

Regards,

Woob Team
""" % (issue.title, issue.id)
        msg = MIMEText(text, 'plain', 'utf-8')
        msg['Subject'] = 'Issue #%s reported' % issue.id
        msg['From'] = 'Woob <woob@woob.tech>'
        msg['To'] = email_to
        s = SMTP('localhost')
        s.sendmail('woob@woob.tech', [email_to], msg.as_string())
        s.quit()

    def do_post(self, line):
        """
        post PROJECT

        Post a new issue.

        If you are not in interactive mode, you can use these parameters:
           --title TITLE
           --assignee ASSIGNEE
           --target-version VERSION
           --category CATEGORY
           --status STATUS
        """
        if not line.strip():
            print('Please give the project name')
            return 1

        project, backend_name = self.parse_id(line, unique_backend=True)

        backend = self.woob.get_backend(backend_name)

        issue = backend.create_issue(project)
        issue.backend = backend.name

        return self.edit_issue(issue, edit=False)

    def complete_edit(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()
        if len(args) == 3:
            return list(dict(self.ISSUE_FIELDS).keys())

    def do_edit(self, line):
        """
        edit ISSUE [KEY [VALUE]]

        Edit an issue.
        If you are not in interactive mode, you can use these parameters:
           --title TITLE
           --assignee ASSIGNEE
           --target-version VERSION
           --category CATEGORY
           --status STATUS
        """
        _id, key, value = self.parse_command_args(line, 3, 1)
        issue = self.get_object(_id, 'get_issue')
        if not issue:
            print('Issue not found: %s' % _id, file=self.stderr)
            return 3

        return self.edit_issue(issue, edit=True)

    def complete_attach(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_attach(self, line):
        """
        attach ISSUE FILENAME

        Attach a file to an issue (Not implemented yet).
        """
        print('Not implemented yet.', file=self.stderr)



# Copyright(C) 2010-2013 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from warnings import warn

from .bcall import CallErrors
from .woob import Woob, WoobBase

__all__ = ['CallErrors', 'Woob', 'WoobBase', 'Weboob', 'WebNip']


class WebNip(WoobBase):
    """Deprecated alias for the freestanding WoobBase application.

    .. deprecated:: 3.0
        Please use :mod:`woob.core.woob.WoobBase` instead.
    """

    def __init__(self, *args, **kwargs):
        warn(
            'WebNip is a deprecated alias and will be removed in Woob 4.0, '
            + 'use WoobBase from woob.core.woob.',
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(*args, **kwargs)


class Weboob(Woob):
    """Deprecated alias for the hosted Woob application.

    .. deprecated:: 3.0
        Please use :mod:`woob.core.woob.Woob` instead.
    """

    def __init__(self, *args, **kwargs):
        warn(
            'Weboob is a deprecated alias and will be removed in Woob 4.0, '
            + 'use Woob from woob.core.woob.',
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(*args, **kwargs)



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import os


__all__ = ['get_javascript']


def get_javascript(name, load_order=('local', 'web'), minified=True):
    if name == 'jquery':
        for src in load_order:
            if src == 'local':
                # try Debian paths
                if minified:
                    filepath = '/usr/share/javascript/jquery/jquery.min.js'
                else:
                    filepath = '/usr/share/javascript/jquery/jquery.js'
                if os.path.exists(filepath):
                    return filepath
            elif src == 'web':
                # return Google-hosted URLs
                if minified:
                    return 'https://ajax.googleapis.com/ajax/libs/jquery/1.4.2/jquery.min.js'
                else:
                    return 'https://ajax.googleapis.com/ajax/libs/jquery/1.4.2/jquery.js'
    elif name == 'tablesorter':
        if 'web' in load_order:
            if minified:
                return 'https://tablesorter.com/jquery.tablesorter.min.js'
            else:
                return 'https://tablesorter.com/jquery.tablesorter.js'
    return None



# Copyright(C) 2010-2016 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from binascii import crc32
import re
from typing import Iterable, List

from woob.capabilities.account import CapCredentialsCheck
from woob.capabilities.base import (
    BaseObject, Capability, Field, StringField, DecimalField, IntField,
    BoolField, UserError, Currency, NotAvailable, EnumField, Enum, empty,
    find_object, NotLoaded, NotLoadedType, NotAvailableType
)
from woob.capabilities.date import DateField
from woob.capabilities.collection import CapCollection
from woob.exceptions import BrowserIncorrectPassword


__all__ = [
    'CapBank', 'BaseAccount', 'Account', 'Loan', 'Transaction', 'AccountNotFound',
    'AccountType', 'AccountOwnership', 'Balance', 'AccountSchemeName', 'TransactionCounterparty',
    'PartyIdentity', 'AccountParty', 'AccountIdentification', 'PartyRole', 'CapAccountCheck',
    'NoAccountsException', 'BalanceType', 'BankTransactionCode',
]


class NoAccountsException(Exception):
    """
    Raised by :meth:`CapBank.iter_accounts()` if we are sure there is no accounts.

    Sometimes we can't parse find accounts on websites but that's a scraping
    error. To attest we know that's a real case, this exception is raised.
    """


class ObjectNotFound(UserError):
    pass


class AccountNotFound(ObjectNotFound):
    """
    Raised when an account is not found.
    """

    def __init__(self, msg: str = 'Account not found'):
        super().__init__(msg)


class BaseAccount(BaseObject, Currency):
    """
    Generic class aiming to be parent of :class:`Recipient` and
    :class:`Account`.
    """
    label =          StringField('Pretty label')
    currency =       StringField('Currency', default=None)
    bank_name =      StringField('Bank Name', mandatory=False)

    # TODO add iban field here

    def __init__(
        self,
        id: str = '0',
        url: str | NotLoadedType | NotAvailableType = NotLoaded
    ):
        super().__init__(id, url)

    @property
    def currency_text(self) -> str:
        return Currency.currency2txt(self.currency)

    @property
    def ban(self) -> str | NotAvailableType:
        """ Bank Account Number part of IBAN"""
        if not self.iban:
            return NotAvailable
        return self.iban[4:]


class AccountType(Enum):
    UNKNOWN          = 0
    CHECKING         = 1
    "Transaction, everyday transactions"
    SAVINGS          = 2
    "Savings/Deposit, can be used for every banking"
    DEPOSIT          = 3
    "Term of Fixed Deposit, has time/amount constraints"
    LOAN             = 4
    "Loan account"
    MARKET           = 5
    "Stock market or other variable investments"
    JOINT            = 6
    "Joint account"
    CARD             = 7
    "Card account"
    LIFE_INSURANCE   = 8
    "Life insurances"
    PEE              = 9
    "Employee savings PEE"
    PERCO            = 10
    "Employee savings PERCO"
    ARTICLE_83       = 11
    "Article 83"
    RSP              = 12
    "Employee savings RSP"
    PEA              = 13
    "Share savings"
    CAPITALISATION   = 14
    "Life Insurance capitalisation"
    PERP             = 15
    "Retirement savings"
    MADELIN          = 16
    "Complementary retirement savings"
    MORTGAGE         = 17
    "Mortgage"
    CONSUMER_CREDIT  = 18
    "Consumer credit"
    REVOLVING_CREDIT = 19
    "Revolving credit"
    PER = 20
    "Pension plan PER"
    REAL_ESTATE = 21
    "Real estate investment such as SCPI, OPCI, SCI"
    CROWDLENDING = 22
    "Crowdlending accounts"
    LDDS = 23
    "LDD/LDDS Livret de développement durable et solidaire"
    PEL = 24
    "Plan épargne logement"
    CSL = 25
    "Compte sur Livret"
    CEL = 26
    "Compte épargne logement"
    CAT = 27
    "Compte à terme"
    LIVRET_A = 28
    "Livret A"
    LIVRET_B = 29
    "Livret B"


class AccountOwnerType:
    """
    Specifies the usage of the account
    """
    PRIVATE = 'PRIV'
    """private personal account"""
    ORGANIZATION = 'ORGA'
    """professional account"""
    ASSOCIATION = 'ASSO'
    """association account"""


class AccountOwnership:
    """
    Relationship between the credentials owner (PSU) and the account
    """
    OWNER = 'owner'
    """The PSU is the account owner"""
    CO_OWNER = 'co-owner'
    """The PSU is the account co-owner"""
    ATTORNEY = 'attorney'
    """The PSU is the account attorney"""


class AccountSchemeName(Enum):
    IBAN = 'iban'
    """IBAN as defined in ISO 13616"""

    BBAN = 'bban'
    """Basic Bank Account Number, represents a country-specific bank account number"""

    SORT_CODE_ACCOUNT_NUMBER = 'sort_code_account_number'
    """Account Identification Number sometimes employed instead of IBAN (e.g.: in UK)"""

    CPAN = 'cpan'
    """Card PAN (masked or plain)"""

    TPAN = 'tpan'
    """Tokenized card PAN issued by a Token Service Provider to obfuscate the real PAN"""

    MPAN = 'mpan'
    """Card PAN where some digits were replaced for security reason"""

    BANK_PARTY_IDENTIFICATION = 'bank_party_identification'
    """
    BankPartyIdentification - Unique and unambiguous assignment made by a specific bank or
    similar financial institution to identify a relationship between the bank and its client.

    Its definition can be found at page 13 of https://www.stet.eu/assets/files/PSD2/1-6-3/api-dsp2-stet-v1.6.3.1-part-2-functional-model.pdf
    """


class AccountManagementType(Enum):
    UNKNOWN = 'unknown'
    CAPITALIZATION = 'capitalization'
    FIXED_FUNDS = 'fixed_funds'
    PROFILED = 'profiled'
    DISCRETIONARY = 'discretionary'
    DELEGATED = 'delegated'
    UNIT_LINKED = 'unit_linked'


class TransactionCounterparty(BaseObject):
    label = StringField('Name of the other stakeholder (Creditor or debtor)', default=None)
    account_scheme_name = EnumField('Type of account Scheme', AccountSchemeName, default=None)
    account_identification = StringField('ID of the account', default=None)
    debtor = BoolField('Type of the counterparty (debtor/creditor/null)', default=None)

    def __repr__(self):
        return f'<label={self.label} debtor={self.debtor} account_scheme_name={self.account_scheme_name} account_identification={self.account_identification}>'


class PartyRole(Enum):
    UNKNOWN = 'unknown'
    HOLDER = 'holder'
    CO_HOLDER = 'co_holder'
    ATTORNEY = 'attorney'
    CUSTODIAN_FOR_MINOR = 'custodian_for_minor'
    LEGAL_GUARDIAN = 'legal_guardian'
    NOMINEE = 'nominee'
    BENEFICIARY = 'beneficiary'
    SUCCESSOR_ON_DEATH = 'successor_on_death'
    TRUSTEE = 'trustee'


class AccountIdentification(BaseObject):
    """
    Defines the identification of a account:
    - scheme_name: Name of the account scheme type
    - identification: ID of the account
    """
    scheme_name = EnumField('Name of the account scheme type', AccountSchemeName, default=None)
    identification = StringField('ID of the account', default=None)

    def __repr__(self):
        return f'<AccountIdentification scheme_name={self.scheme_name} identification={self.identification}>'


class PartyIdentity(BaseObject):
    """
    Defines the identity of a party:
    - full_name: Full name of the party
    - role: Role of the party
    - is_user: Defines the link between the party and the connected PSU
    """
    ROLE_UNKNOWN = PartyRole.UNKNOWN
    ROLE_HOLDER = PartyRole.HOLDER
    ROLE_CO_HOLDER = PartyRole.CO_HOLDER
    ROLE_ATTORNEY = PartyRole.ATTORNEY
    ROLE_CUSTODIAN_FOR_MINOR = PartyRole.CUSTODIAN_FOR_MINOR
    ROLE_LEGAL_GUARDIAN = PartyRole.LEGAL_GUARDIAN
    ROLE_NOMINEE = PartyRole.NOMINEE
    ROLE_BENEFICIARY = PartyRole.BENEFICIARY
    ROLE_SUCCESSOR_ON_DEATH = PartyRole.SUCCESSOR_ON_DEATH
    ROLE_TRUSTEE = PartyRole.TRUSTEE

    full_name = StringField('Full name of the party.', default=None)
    is_user = BoolField('Is the party the connected PSU?', default=None)
    role = EnumField('Role of the party.', PartyRole, default=ROLE_UNKNOWN)

    def __repr__(self):
        return f'<PartyIdentity full_name={self.full_name} role={self.role} is_user={self.is_user}>'


class AccountParty(BaseObject):
    """
    Defines all the information related to an account party:
    - party_identities: list of PartyIdentity elements
    - account_identifications : list of AccountIdentification elements
    """
    party_identities = Field('Identities of the account party', list, default=[])
    account_identifications = Field('Identification information of the account', list, default=[])

    def __repr__(self):
        return f'<AccountParty party_identities={self.party_identities} account_identifications={self.account_identifications}>'


class Account(BaseAccount):
    """
    Bank account.
    """
    TYPE_UNKNOWN          = AccountType.UNKNOWN
    TYPE_CHECKING         = AccountType.CHECKING
    TYPE_SAVINGS          = AccountType.SAVINGS
    TYPE_DEPOSIT          = AccountType.DEPOSIT
    TYPE_LOAN             = AccountType.LOAN
    TYPE_MARKET           = AccountType.MARKET
    TYPE_JOINT            = AccountType.JOINT
    TYPE_CARD             = AccountType.CARD
    TYPE_LIFE_INSURANCE   = AccountType.LIFE_INSURANCE
    TYPE_PEE              = AccountType.PEE
    TYPE_PERCO            = AccountType.PERCO
    TYPE_ARTICLE_83       = AccountType.ARTICLE_83
    TYPE_RSP              = AccountType.RSP
    TYPE_PEA              = AccountType.PEA
    TYPE_CAPITALISATION   = AccountType.CAPITALISATION
    TYPE_PERP             = AccountType.PERP
    TYPE_MADELIN          = AccountType.MADELIN
    TYPE_MORTGAGE         = AccountType.MORTGAGE
    TYPE_CONSUMER_CREDIT  = AccountType.CONSUMER_CREDIT
    TYPE_REVOLVING_CREDIT = AccountType.REVOLVING_CREDIT
    TYPE_PER              = AccountType.PER
    TYPE_REAL_ESTATE      = AccountType.REAL_ESTATE
    TYPE_CROWDLENDING     = AccountType.CROWDLENDING
    TYPE_LDDS             = AccountType.LDDS
    TYPE_PEL              = AccountType.PEL
    TYPE_CSL              = AccountType.CSL
    TYPE_CEL              = AccountType.CEL
    TYPE_CAT              = AccountType.CAT
    TYPE_LIVRET_A         = AccountType.LIVRET_A
    TYPE_LIVRET_B         = AccountType.LIVRET_B

    type =      EnumField('Type of account', AccountType, default=TYPE_UNKNOWN)
    owner_type = StringField('Usage of account')  # cf AccountOwnerType class
    balance =   DecimalField('Balance on this bank account')
    coming =    DecimalField('Sum of coming movements')
    iban =      StringField('International Bank Account Number', mandatory=False)
    ownership = StringField('Relationship between the credentials owner (PSU) and the account')  # cf AccountOwnership class

    # card attributes
    paydate =   DateField('For credit cards. When next payment is due.')
    paymin =    DecimalField('For credit cards. Minimal payment due.')
    cardlimit = DecimalField('For credit cards. Credit limit.')

    number =    StringField('Shown by the bank to identify your account ie XXXXX7489')

    # Wealth accounts (market, life insurance...)
    valuation_diff = DecimalField('+/- values total')
    valuation_diff_ratio = DecimalField('+/- values ratio')
    management_type = EnumField('Management type of account', AccountManagementType, default=None)

    # Employee savings (PERP, PERCO, Article 83...)
    company_name = StringField('Name of the company of the stock - only for employee savings')

    # parent account
    #  - A checking account parent of a card account
    #  - A checking account parent of a recurring loan account
    #  - An investment account parent of a liquidity account
    #  - ...
    parent = Field('Parent account', BaseAccount)

    opening_date = DateField('Date when the account contract was created on the bank')

    all_balances = Field('List of balances', list, default=[])

    party = Field('Party associated to the account', AccountParty, default=None)

    def __repr__(self):
        return "<%s id=%r label=%r>" % (type(self).__name__, self.id, self.label)

    # compatibility alias
    @property
    def valuation_diff_percent(self):
        return self.valuation_diff_ratio

    @valuation_diff_percent.setter
    def valuation_diff_percent(self, value):
        self.valuation_diff_ratio = value


class BalanceType(Enum):
    CLOSING = 1
    """Current balance of the account"""
    PENDING = 2
    """Forecast balance of the account"""


class Balance(BaseObject):
    """
    Object made to receive balance on one Account
    """

    amount = DecimalField('Amount on this balance')
    type = EnumField('Type of balance', BalanceType)
    currency = StringField('Currency')
    reference_date = DateField('date of the balance')
    last_update = DateField('Last time balance was updated')
    credit_included = BoolField('If factoring is included in balance', default=False)
    label = StringField('Bank name of the balance')
    calculated = BoolField('If computation has been made on the balance', default=False)

    def __repr__(self):
        # Ex: '< Balance: label="Solde en Valeur" amount=972.94 type=1 credit_included=False reference_date=2023-06-09 >'
        return ' '.join((
            '<',
            f'{type(self).__name__}:',
            f'label="{self.label}"',
            f'amount={self.amount}',
            f'type={self.type}',
            f'credit_included={self.credit_included}',
            f'reference_date={self.reference_date}',
            '>'
        ))


class Loan(Account):
    """
    Account type dedicated to loans and credits.
    """

    name = StringField('Person name')
    account_label = StringField('Label of the debited account')
    insurance_label = StringField('Label of the insurance')

    total_amount = DecimalField('Total amount loaned')
    available_amount = DecimalField('Amount available') # only makes sense for revolving credit
    used_amount = DecimalField('Amount already used') # only makes sense for revolving credit

    insurance_amount = DecimalField("Amount of the loan's insurance")
    insurance_rate = DecimalField("Rate of the loan's insurance")

    subscription_date = DateField('Date of subscription of the loan')
    maturity_date = DateField('Estimated end date of the loan')
    start_repayment_date = DateField('Date of start repayment of the loan')
    deferred = BoolField('If loan is deferred')
    duration = IntField('Duration of the loan given in months')
    rate = DecimalField('Monthly rate of the loan')

    nb_payments_left = IntField('Number of payments still due')
    nb_payments_done = IntField('Number of payments already done')
    nb_payments_total = IntField('Number total of payments')

    last_payment_amount = DecimalField('Amount of the last payment done')
    last_payment_date = DateField('Date of the last payment done')
    next_payment_amount = DecimalField('Amount of next payment')
    next_payment_date = DateField('Date of the next payment')


class TransactionType(Enum):
    UNKNOWN       = 0
    TRANSFER      = 1
    ORDER         = 2
    CHECK         = 3
    DEPOSIT       = 4
    PAYBACK       = 5
    WITHDRAWAL    = 6
    CARD          = 7
    LOAN_PAYMENT  = 8
    BANK          = 9
    CASH_DEPOSIT  = 10
    CARD_SUMMARY  = 11
    DEFERRED_CARD = 12
    INSTANT       = 13
    MARKET_ORDER  = 14
    MARKET_FEE    = 15
    ARBITRAGE     = 16
    PROFIT        = 17


class BankTransactionCode(BaseObject):
    """
    Object dedicating to bank transaction codes
    It follows the ISO20022 standards.
    See https://www.iso20022.org/catalogue-messages/additional-content-messages/external-code-sets
    """
    domain = StringField('Domain of the transaction')
    family = StringField('Family of the transaction')
    sub_family = StringField('Sub-family of the transaction')


class Transaction(BaseObject):
    """
    Bank transaction.
    """
    TYPE_UNKNOWN       = TransactionType.UNKNOWN
    TYPE_TRANSFER      = TransactionType.TRANSFER
    TYPE_ORDER         = TransactionType.ORDER
    TYPE_CHECK         = TransactionType.CHECK
    TYPE_DEPOSIT       = TransactionType.DEPOSIT
    TYPE_PAYBACK       = TransactionType.PAYBACK
    TYPE_WITHDRAWAL    = TransactionType.WITHDRAWAL
    TYPE_CARD          = TransactionType.CARD
    TYPE_LOAN_PAYMENT  = TransactionType.LOAN_PAYMENT
    TYPE_BANK          = TransactionType.BANK
    TYPE_CASH_DEPOSIT  = TransactionType.CASH_DEPOSIT
    TYPE_CARD_SUMMARY  = TransactionType.CARD_SUMMARY
    TYPE_DEFERRED_CARD = TransactionType.DEFERRED_CARD
    TYPE_INSTANT       = TransactionType.INSTANT
    TYPE_MARKET_ORDER  = TransactionType.MARKET_ORDER
    TYPE_MARKET_FEE    = TransactionType.MARKET_FEE
    TYPE_ARBITRAGE     = TransactionType.ARBITRAGE
    TYPE_PROFIT        = TransactionType.PROFIT

    date =      DateField('Debit date on the bank statement')
    rdate =     DateField('Real date, when the payment has been made; usually extracted from the label or from credit card info')
    vdate =     DateField('Value date, or accounting date; usually for professional accounts')
    bdate =     DateField('Bank date, when the transaction appear on website (usually extracted from column date)')
    type =      EnumField('Type of transaction, use TYPE_* constants', TransactionType, default=TYPE_UNKNOWN)
    raw =       StringField('Raw label of the transaction')
    category =  StringField('Category of the transaction')
    label =     StringField('Pretty label')
    amount = DecimalField('Net amount of the transaction, used to compute account balance')
    coming = BoolField('True if the transaction is not yet booked')

    card =              StringField('Card number (if any)')
    commission =        DecimalField('Commission part on the transaction (in account currency)')
    gross_amount = DecimalField('Amount of the transaction without the commission')

    # International
    original_amount = DecimalField('Original net amount (in another currency)')
    original_currency = StringField('Currency of the original amount')
    country =           StringField('Country of transaction')

    original_commission =          DecimalField('Original commission (in another currency)')
    original_commission_currency = StringField('Currency of the original commission')
    original_gross_amount = DecimalField('Original gross amount (in another currency)')

    attachments =       Field('List of files attached to the transaction', list)
    # Financial arbitrations
    investments =       Field('List of investments related to the transaction', list, default=[])

    counterparty = Field('Counterparty of transaction', TransactionCounterparty)

    bank_transaction_code = Field('Bank transaction code of transaction', BankTransactionCode)

    def __repr__(self):
        return "<Transaction date=%r label=%r amount=%r>" % (self.date, self.label, self.amount)

    def unique_id(self, seen: set | None = None, account_id: str | None = None) -> str:
        """
        Get an unique ID for the transaction based on date, amount and raw.

        :param seen: if given, the method uses this set as a cache to
                     prevent several transactions with the same values to have the same
                     unique ID.
        :type seen: :class:`set`
        :param account_id: if given, add the account ID in data used to create
                           the unique ID. Can be useful if you want your ID to be unique across
                           several accounts.
        :type account_id: :class:`str`
        :returns: an unique ID encoded in 8 length hexadecimal string (for example ``'a64e1bc9'``)
        :rtype: :class:`str`
        """
        crc = crc32(str(self.date).encode('utf-8'))
        crc = crc32(str(self.amount).encode('utf-8'), crc)
        if not empty(self.raw):
            label = self.raw
        else:
            label = self.label

        crc = crc32(re.sub('[ ]+', ' ', label).encode("utf-8"), crc)

        if account_id is not None:
            crc = crc32(str(account_id).encode('utf-8'), crc)

        if seen is not None:
            while crc in seen:
                crc = crc32(b"*", crc)

            seen.add(crc)

        return "%08x" % (crc & 0xffffffff)


class CapBank(CapCollection, CapCredentialsCheck):
    """
    Capability of bank websites to see accounts and transactions.
    """

    def check_credentials(self) -> bool:
        """
        Check that the given credentials are correct by trying to login.

        The default implementation of this method check if the class using this capability
        has a browser, execute its do_login if it has one and then see if no error pertaining to the creds is raised.
        If any other unexpected error occurs, we don't know whether the creds are correct or not.
        """
        # TODO move this in a specific capability
        if getattr(self, 'BROWSER', None) is None:
            raise NotImplementedError()

        try:
            self.browser.do_login()
        except BrowserIncorrectPassword:
            return False

        return True

    def iter_resources(self, objs: List[BaseObject], split_path: List[str]) -> Iterable[BaseObject]:
        """
        Iter resources.

        Default implementation of this method is to return on top-level
        all accounts (by calling :func:`iter_accounts`).

        :param objs: type of objects to get
        :type objs: tuple[:class:`BaseObject`]
        :param split_path: path to discover
        :type split_path: :class:`list`
        :rtype: iter[:class:`BaseObject`]
        """
        if Account in objs:
            self._restrict_level(split_path)

            yield from self.iter_accounts()

    def iter_accounts(self) -> Iterable[Account]:
        """
        Iter accounts.

        :rtype: iter[:class:`Account`]
        """
        raise NotImplementedError()

    def get_account(self, id: str) -> Account | None:
        """
        Get an account from its ID.

        :param id: ID of the account
        :type id: :class:`str`
        :rtype: :class:`Account`
        :raises: :class:`AccountNotFound`
        """
        return find_object(self.iter_accounts(), id=id, error=AccountNotFound)

    def iter_history(self, account: Account) -> Iterable[Transaction]:
        """
        Iter history of transactions of a specific account.

        :param account: account to get history
        :type account: :class:`Account`
        :rtype: iter[:class:`Transaction`]
        :raises: :class:`AccountNotFound`
        """
        return self.iter_transactions(account, with_coming=False)

    def iter_coming(self, account: Account) -> Iterable[Transaction]:
        """
        Iter coming transactions of a specific account.

        :param account: account to get coming transactions
        :type account: :class:`Account`
        :rtype: iter[:class:`Transaction`]
        :raises: :class:`AccountNotFound`
        """
        return self.iter_transactions(account, with_history=False)

    def iter_transactions(
        self, account: Account, *, with_history: bool = True, with_coming: bool = True
    ) -> Iterable[Transaction]:
        """
        Iter all transactions (history and coming) of a specific account.

        :param account: account to get transactions
        :param with_history: if False, booked transactions will not be returned
        :param with_coming: if False, coming transactions will not be returned
        :type account: :class:`Account`
        :rtype: iter[:class:`Transaction`]
        :raises: :class:`AccountNotFound`
        """
        raise NotImplementedError()


class CapAccountCheck(Capability):
    """
    Capability to get accounts parties information.

    The expected structure is the following:
        - AccountParty object
            * party_identities (list of PartyIdentity elements):
                * full name
                * role
                * is_user
            * account_identifications (list of type AccountIdentification elements):
                * scheme name
                * identification
    """



# Copyright(C) 2017 Jonathan Schmidt
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal
import re

from woob.capabilities.base import NotAvailable
from woob.capabilities.bank.wealth import Investment
from woob.browser.filters.base import Filter, FilterError, debug
from woob.tools.log import getLogger

def is_isin_valid(isin):
    """
    Méthode générale
    Table de conversion des lettres en chiffres
    A=10    B=11    C=12    D=13    E=14    F=15    G=16    H=17    I=18
    J=19    K=20    L=21    M=22    N=23    O=24    P=25    Q=26    R=27
    S=28    T=29    U=30    V=31    W=32    X=33    Y=34    Z=35

    1 - Mettre de côté la clé, qui servira de référence à la fin de la vérification.
    2 - Convertir toutes les lettres en nombres via la table de conversion ci-contre. Si le nombre obtenu est supérieur ou égal à 10, prendre les deux chiffres du nombre séparément (exemple : 27 devient 2 et 7).
    3 - Pour chaque chiffre, multiplier sa valeur par deux si sa position est impaire en partant de la droite. Si le nombre obtenu est supérieur ou égal à 10, garder les deux chiffres du nombre séparément (exemple : 14 devient 1 et 4).
    4 - Faire la somme de tous les chiffres.
    5 - Soustraire cette somme de la dizaine supérieure ou égale la plus proche (exemples : si la somme vaut 22, la dizaine « supérieure ou égale » est 30, et la clé vaut donc 8 ; si la somme vaut 30, la dizaine « supérieure ou égale » est 30, et la clé vaut 0 ; si la somme vaut 31, la dizaine « supérieure ou égale » est 40, et la clé vaut 9).
    6 - Comparer la valeur obtenue à la clé mise initialement de côté.

    Étapes 1 et 2 :
    F R 0 0 0 3 5 0 0 0 0 (+ 8 : clé)
    15 27 0 0 0 3 5 0 0 0 0

    Étape 3 : le traitement se fait sur des chiffres
    1 5 2 7 0 0 0 3 5 0 0 0 0
    I P I P I P I P I P I P I : position en partant de la droite (P = Pair, I = Impair)
    2 1 2 1 2 1 2 1 2 1 2 1 2 : coefficient multiplicateur

    2 5 4 7 0 0 0 3 10 0 0 0 0 : résultat

    Étape 4 :
    2 + 5 + 4 + 7 + 0 + 0 + 0 + 3 + (1 + 0)+ 0 + 0 + 0 + 0 = 22

    Étapes 5 et 6 : 30 - 22 = 8 (valeur de la clé)
    """

    if not isinstance(isin, str):
        return False
    if not re.match(r'^[A-Z]{2}[A-Z0-9]{9}\d$', isin):
        return False

    isin_in_digits = ''.join(str(ord(x) - ord('A') + 10) if not x.isdigit() else x for x in isin[:-1])
    key = isin[-1:]
    result = ''
    for k, val in enumerate(isin_in_digits[::-1], start=1):
        if k % 2 == 0:
            result = ''.join((result, val))
        else:
            result = ''.join((result, str(int(val)*2)))
    return str(sum(int(x) for x in result) + int(key))[-1] == '0'


def create_french_liquidity(valuation):
    """
    Automatically fills a liquidity investment with label, code and code_type.
    """
    if isinstance(valuation, Decimal) and valuation < 0:
        getLogger('%s.create_French_liquidity' % __name__).warning("Liquidity has a negative value")
    liquidity = Investment()
    liquidity.label = "Liquidités"
    liquidity.code = "XX-liquidity"
    liquidity.code_type = NotAvailable
    liquidity.valuation = valuation
    return liquidity


# These filters can be used to set Investment.code
# and Investment.code_type without having to declare
# obj_code() and obj_code_type() methods in each module

class FormatError(FilterError):
    pass


class IsinCode(Filter):
    """
    Returns the input only if it is a valid ISIN code.
    """
    @debug()
    def filter(self, code):
        if is_isin_valid(code):
            return code
        return self.default_or_raise(FormatError('%r is not a valid ISIN code, no default value was set.' % code))


class IsinType(Filter):
    """
    Returns Investment.CODE_TYPE_ISIN if the input is a valid ISIN code.
    """
    @debug()
    def filter(self, code):
        if is_isin_valid(code):
            return Investment.CODE_TYPE_ISIN
        return NotAvailable



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
import shlex
import subprocess
from shutil import which

from woob.core.bcall import CallErrors
from woob.capabilities.content import CapContent, Revision, Content
from woob.tools.application.repl import ReplApplication, defaultcount


__all__ = ['AppContentEdit']


class AppContentEdit(ReplApplication):
    APPNAME = 'contentedit'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to display and edit contents on various websites."
    SHORT_DESCRIPTION = "manage websites content"
    CAPS = CapContent

    def do_edit(self, line):
        """
        edit ID [ID...]

        Edit a content with $EDITOR, then push it on the website.
        """
        contents = []
        for id in line.split():
            _id, backend_name = self.parse_id(id, unique_backend=True)
            backend_names = (backend_name,) if backend_name is not None else self.enabled_backends

            contents += [content for content in self.do('get_content', _id, backends=backend_names) if content]

        if len(contents) == 0:
            print('No contents found', file=self.stderr)
            return 3

        if self.stdin.isatty():
            paths = {}
            for content in contents:
                tmpdir = os.path.join(tempfile.gettempdir(), "woob")
                if not os.path.isdir(tmpdir):
                    os.makedirs(tmpdir)
                with tempfile.NamedTemporaryFile('w+t', prefix='%s_' % content.id.replace(os.path.sep, '_'), dir=tmpdir, delete=False) as f:
                    if content.content is None:
                        content.content = ''
                    f.write(content.content)
                paths[f.name] = content

            params = []
            editor = os.environ.get('EDITOR', 'vim')
            # check cases where /usr/bin/vi is a symlink to vim
            if 'vim' in (os.path.basename(editor), os.path.basename(os.path.realpath(which(editor) or '/')).replace('.nox', '')):
                params = ['-p']
            subprocess.call([editor, *params, *paths])

            for path, content in paths.items():
                with open(path, 'r') as f:
                    try:
                        data = f.read()
                    except UnicodeError:
                        pass
                if content.content != data:
                    content.content = data
                else:
                    contents.remove(content)

            if len(contents) == 0:
                print('No changes. Abort.', file=self.stderr)
                return 1

            print('Contents changed:\n%s' % ('\n'.join(' * %s' % content.id for content in contents)))

            message = self.ask('Enter a commit message', default='')
            minor = self.ask('Is this a minor edit?', default=False)
            if not self.ask('Do you want to push?', default=True):
                return

            errors = CallErrors([])
            for content in contents:
                path = [path for path, c in paths.items() if c == content][0]
                self.stdout.write('Pushing %s...' % content.id)
                self.stdout.flush()
                try:
                    self.do('push_content', content, message, minor=minor, backends=[content.backend]).wait()
                except CallErrors as e:
                    errors.errors += e.errors
                    self.stdout.write(' error (content saved in %s)\n' % path)
                else:
                    self.stdout.write(' done\n')
                    os.unlink(path)
        else:
            # stdin is not a tty

            if len(contents) != 1:
                print("Multiple ids not supported with pipe", file=self.stderr)
                return 2

            message, minor = '', False
            data = self.stdin.read()
            contents[0].content = data

            errors = CallErrors([])
            for content in contents:
                self.stdout.write('Pushing %s...' % content.id)
                self.stdout.flush()
                try:
                    self.do('push_content', content, message, minor=minor, backends=[content.backend]).wait()
                except CallErrors as e:
                    errors.errors += e.errors
                    self.stdout.write(' error\n')
                else:
                    self.stdout.write(' done\n')

        if len(errors.errors) > 0:
            raise errors

    def do_create(self, line):
        """
        create TITLE BACKEND
        """
        args = shlex.split(line)
        title, backend = args

        if self.stdin.isatty():
            editor = os.environ.get('EDITOR', 'vi')
            with tempfile.NamedTemporaryFile('w+t', suffix='.md') as fd:
                subprocess.call([editor, fd.name])
                data = fd.read()
        else:
            data = self.stdin.read()

        content = Content()
        content.title = args[0]
        content.content = data
        content.backend = backend
        content = next(iter(self.do('push_content', content, message='', minor=False, backends=[content.backend]))) or content
        if content.url:
            print('Pushed to', content.url, file=self.stdout)

    @defaultcount(10)
    def do_log(self, line):
        """
        log ID

        Display log of a page
        """
        if not line:
            print('Error: please give a page ID', file=self.stderr)
            return 2

        _id, backend_name = self.parse_id(line)
        backend_names = (backend_name,) if backend_name is not None else self.enabled_backends

        self.start_format()
        for revision in self.do('iter_revisions', _id, backends=backend_names):
            self.format(revision)

    def do_get(self, line):
        """
        get ID [-r revision]

        Get page contents
        """
        if not line:
            print('Error: please give a page ID', file=self.stderr)
            return 2

        _part_line = line.strip().split(' ')
        revision = None
        if '-r' in _part_line:
            r_index = _part_line.index('-r')
            if len(_part_line) -1 > r_index:
                revision = Revision(_part_line[r_index+1])
                _part_line.remove(revision.id)
            _part_line.remove('-r')

            if not _part_line:
                print('Error: please give a page ID', file=self.stderr)
                return 2

        _id, backend_name = self.parse_id(" ".join(_part_line))

        backend_names = (backend_name,) if backend_name is not None else self.enabled_backends

        output = self.stdout
        for contents in [content for content in self.do('get_content', _id, revision, backends=backend_names) if content]:
            output.write(contents.content)

        # add a newline unless we are writing
        # in a file or in a pipe
        if output.isatty():
            output.write('\n')



# Copyright(C) 2010-2021  Christophe Benz, Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import atexit
import logging
import os
import re
import shlex
import signal
from cmd import Cmd
from collections import OrderedDict, deque
from datetime import datetime
from optparse import IndentedHelpFormatter, OptionGroup, OptionParser

import yaml

from woob.capabilities.base import BaseObject, FieldNotFound, UserError, empty
from woob.capabilities.collection import BaseCollection, CapCollection, Collection, CollectionNotFound
from woob.core import CallErrors
from woob.exceptions import BrowserQuestion, BrowserRedirect, DecoupledValidation
from woob.tools.application.formatters.iformatter import MandatoryFieldsNotFound
from woob.tools.path import WorkingPath

from .console import BackendNotGiven, ConsoleApplication
from .formatters.load import FormatterLoadError, FormattersLoader
from .results import ResultsCondition, ResultsConditionError
from .pretty import colored


__all__ = [
    'NotEnoughArguments', 'TooManyArguments', 'ArgSyntaxError',
    'ReplApplication'
]


class NotEnoughArguments(Exception):
    pass


class TooManyArguments(Exception):
    pass


class ArgSyntaxError(Exception):
    pass


class ReplOptionParser(OptionParser):
    def format_option_help(self, formatter=None):
        if not formatter:
            formatter = self.formatter

        return '%s\n%s' % (formatter.format_commands(self.commands),
                           OptionParser.format_option_help(self, formatter))


class ReplOptionFormatter(IndentedHelpFormatter):
    def format_commands(self, commands):
        s = ''
        for section, cmds in commands.items():
            if len(cmds) == 0:
                continue
            if len(s) > 0:
                s += '\n'
            s += '%s Commands:\n' % section
            for c in cmds:
                c = c.split('\n')[0]
                s += '    %s\n' % c
        return s


def defaultcount(default_count=10):
    def deco(f):
        def inner(self, *args, **kwargs):
            oldvalue = self.options.count
            if self._is_default_count:
                self.options.count = default_count

            try:
                return f(self, *args, **kwargs)
            finally:
                self.options.count = oldvalue

        inner.__doc__ = f.__doc__
        assert inner.__doc__ is not None, "A command must have a docstring"
        inner.__doc__ += '\nDefault is limited to %s results.' % default_count

        return inner
    return deco


class ReplApplication(ConsoleApplication, Cmd):
    """
    Base application class for Repl applications.
    """

    SYNOPSIS =  'Usage: %prog [-dqv] [-b backends] [-cnfs] [command [arguments..]]\n'
    SYNOPSIS += '       %prog [--help] [--version]'
    DISABLE_REPL = False

    EXTRA_FORMATTERS = {}
    DEFAULT_FORMATTER = 'multiline'
    COMMANDS_FORMATTERS = {}

    COLLECTION_OBJECTS = tuple()
    """Objects to allow in do_ls / do_cd"""

    woob_commands = {'backends', 'condition', 'count', 'formatter', 'logging', 'select', 'quit', 'ls', 'cd', 'storage'}
    hidden_commands = {'EOF'}

    def __init__(self):
        super().__init__(ReplOptionParser(self.SYNOPSIS, version=self._get_optparse_version()))

        copyright = self.COPYRIGHT.replace('YEAR', '%d' % datetime.today().year)
        self.intro = '\n'.join(('Welcome to %s v%s' % (colored(self.APPNAME, attrs=['bold']), self.VERSION),
                                '',
                                copyright,
                                'This program is free software: you can redistribute it and/or modify',
                                'it under the terms of the GNU Lesser General Public License as published by',
                                'the Free Software Foundation, either version 3 of the License, or',
                                '(at your option) any later version.',
                                '',
                                'Type "help" to display available commands.',
                                '',
                               ))
        self.formatters_loader = FormattersLoader()
        for key, klass in self.EXTRA_FORMATTERS.items():
            self.formatters_loader.register_formatter(key, klass)
        self.formatter = None
        self.commands_formatters = self.COMMANDS_FORMATTERS.copy()

        commands_help = self.get_commands_doc()
        self._parser.commands = commands_help
        self._parser.formatter = ReplOptionFormatter()

        results_options = OptionGroup(self._parser, 'Results Options')
        results_options.add_option('-c', '--condition', help='filter result items to display given a boolean expression. See CONDITION section for the syntax')
        results_options.add_option('-n', '--count', type='int',
                                   help='limit number of results (from each backends)')
        results_options.add_option('-s', '--select', help='select result item keys to display (comma separated)')
        self._parser.add_option_group(results_options)

        formatting_options = OptionGroup(self._parser, 'Formatting Options')
        available_formatters = self.formatters_loader.get_available_formatters()
        formatting_options.add_option('-f', '--formatter', choices=available_formatters,
                                      help='select output formatter (%s)' % ', '.join(available_formatters))
        formatting_options.add_option('--no-header', dest='no_header', action='store_true', help='do not display header')
        formatting_options.add_option('--no-keys', dest='no_keys', action='store_true', help='do not display item keys')
        formatting_options.add_option('-O', '--outfile', dest='outfile', help='file to export result')
        self._parser.add_option_group(formatting_options)

        self._interactive = False
        self.working_path = WorkingPath()
        self._change_prompt()

    @property
    def interactive(self):
        return self._interactive

    def _change_prompt(self):
        self.objects = []
        self.collections = []
        # XXX can't use bold prompt because:
        # 1. it causes problems when trying to get history (lines don't start
        #    at the right place).
        # 2. when typing a line longer than term width, cursor goes at start
        #    of the same line instead of new line.
        #self.prompt = self.BOLD + '%s> ' % self.APPNAME + self.NC
        if len(self.working_path.get()):
            wp_enc = str(self.working_path)
            self.prompt = '%s:%s> ' % (self.APPNAME, wp_enc)
        else:
            self.prompt = '%s> ' % (self.APPNAME)

    def change_path(self, split_path):
        self.working_path.location(split_path)
        self._change_prompt()

    def add_object(self, obj):
        self.objects.append(obj)

    def _complete_object(self):
        return [obj.fullid for obj in self.objects]

    def parse_id(self, id, unique_backend=False):
        if self.interactive:
            try:
                obj = self.objects[int(id) - 1]
            except (IndexError, ValueError):
                # Try to find a shortcut in the cache
                for obj in self.objects:
                    if id in obj.id:
                        id = obj.fullid
                        break
            else:
                if isinstance(obj, BaseObject):
                    id = obj.fullid
        try:
            return super().parse_id(id, unique_backend)
        except BackendNotGiven as e:
            backend_name = None
            while not backend_name:
                print('This command works with an unique backend. Availables:')
                for index, (name, backend) in enumerate(e.backends):
                    print(
                        '%s) %s%s' % (
                            colored('%2d' % (index + 1), 'white', attrs=['bold']),
                            colored('%-20s' % name, 'magenta', attrs=['bold']),
                            colored(backend.DESCRIPTION, 'green')
                        )
                    )


                i = self.ask('Select a backend to proceed with "%s"' % id)
                if not i.isdigit():
                    if i not in dict(e.backends):
                        print('Error: %s is not a valid backend' % i, file=self.stderr)
                        continue
                    backend_name = i
                else:
                    i = int(i)
                    if i < 0 or i > len(e.backends):
                        print('Error: %s is not a valid choice' % i, file=self.stderr)
                        continue
                    backend_name = e.backends[i-1][0]

            return id, backend_name

    def get_object(self, _id, method, fields=None, caps=None):
        if self.interactive:
            try:
                obj = self.objects[int(_id) - 1]
            except (IndexError, ValueError):
                pass
            else:
                try:
                    backend = self.woob.get_backend(obj.backend)
                    actual_method = getattr(backend, method, None)
                    if actual_method is None:
                        return None
                    else:
                        if callable(actual_method):
                            obj, = self.do('fillobj', obj, fields, backends=backend)
                            return obj
                        else:
                            return None
                except UserError as e:
                    self.bcall_error_handler(backend, e, '')

        _id, backend_name = self.parse_id(_id)
        kargs = {}
        if caps is not None:
            kargs = {'caps': caps}
        backend_names = (backend_name,) if backend_name is not None else self.enabled_backends

        # if backend's service returns several objects, try to find the one
        # with wanted ID. If not found, get the last not None object.
        obj = None

        # remove backends that do not have the required method
        new_backend_names = []
        for backend in backend_names:
            if isinstance(backend, str):
                actual_backend = self.woob.get_backend(backend)
            else:
                actual_backend = backend
            if getattr(actual_backend, method, None) is not None:
                new_backend_names.append(backend)
        backend_names = tuple(new_backend_names)
        try:
            for objiter in self.do(method, _id, backends=backend_names, fields=fields, **kargs):
                if objiter:
                    obj = objiter
                    if objiter.id == _id:
                        return obj
        except CallErrors as e:
            if obj is not None:
                self.bcall_errors_handler(e)
            else:
                raise

        return obj

    def get_object_list(self, method=None, *args, **kwargs):
        # return cache if not empty
        if len(self.objects) > 0:
            return self.objects
        elif method is not None:
            kwargs['backends'] = self.enabled_backends
            for _object in self.do(method, *args, **kwargs):
                self.add_object(_object)
            return self.objects
        # XXX: what can we do without method?
        return tuple()

    def unload_backends(self, *args, **kwargs):
        self.objects = []
        self.collections = []
        return super().unload_backends(*args, **kwargs)

    def load_backends(self, *args, **kwargs):
        self.objects = []
        self.collections = []
        return super().load_backends(*args, **kwargs)

    def main(self, argv):
        cmd_args = argv[1:]
        if cmd_args:
            cmd_line = ' '.join(cmd_args)
            cmds = cmd_line.split(';')
            for cmd in cmds:
                ret = self.onecmd(cmd)
                if ret:
                    return ret
        elif self.DISABLE_REPL:
            self._parser.print_help()
            self._parser.exit()
        else:
            try:
                import readline
            except ImportError:
                pass
            else:
                # Remove '-' from delims
                readline.set_completer_delims(readline.get_completer_delims().replace('-', ''))

                history_filepath = os.path.join(self.woob.workdir, '%s_history' % self.APPNAME)
                if self.OLD_APPNAME:
                    # compatibility for old, non-woob names
                    history_filepath = self._get_preferred_path(
                        history_filepath, os.path.join(self.woob.workdir, '%s_history' % self.OLD_APPNAME)
                    )
                try:
                    readline.read_history_file(history_filepath)
                except IOError:
                    pass

                def savehist():
                    readline.write_history_file(history_filepath)
                atexit.register(savehist)

            self.intro += '\nLoaded backends: %s\n' % ', '.join(sorted(backend.name for backend in self.woob.iter_backends()))
            self._interactive = True
            self.cmdloop()

    def do(self, function, *args, **kwargs):
        """
        Call Woob.do(), passing count and selected fields given by user.
        """
        backends = kwargs.pop('backends', None)
        if backends is None:
            kwargs['backends'] = []
            for backend in self.enabled_backends:
                actual_function = getattr(backend, function, None) if isinstance(function, str) else function

                if callable(actual_function):
                    kwargs['backends'].append(backend)
        else:
            kwargs['backends'] = backends
        fields = kwargs.pop('fields', self.selected_fields)
        if not fields and fields != []:
            fields = self.selected_fields

        fields = self.parse_fields(fields)

        if fields and self.formatter.MANDATORY_FIELDS is not None:
            missing_fields = set(self.formatter.MANDATORY_FIELDS) - set(fields)
            # If a mandatory field is not selected, do not use the customized formatter
            if missing_fields:
                print('Warning: you do not select enough mandatory fields for the formatter. Fallback to another. Hint: use option -f', file=self.stderr)
                self.formatter = self.formatters_loader.build_formatter(ReplApplication.DEFAULT_FORMATTER)

        if self.formatter.DISPLAYED_FIELDS is not None:
            if fields is None:
                missing_fields = True
            else:
                missing_fields = set(fields) - set(self.formatter.DISPLAYED_FIELDS + self.formatter.MANDATORY_FIELDS)
            # If a selected field is not displayed, do not use the customized formatter
            if missing_fields:
                print('Warning: some selected fields will not be displayed by the formatter. Fallback to another. Hint: use option -f', file=self.stderr)
                self.formatter = self.formatters_loader.build_formatter(ReplApplication.DEFAULT_FORMATTER)

        return self.woob.do(self._do_complete, self.options.count, fields, function, *args, **kwargs)

    # TODO make _do_and_retry and _do_and_retry_wait public
    # TODO (bis): make this comportment as the default of 'do', with support of 'wait()' call on the returned object.
    def _do_and_retry_wait(self, *args, **kwargs):
        deque(self._do_and_retry(*args, **kwargs), maxlen=0)

    def _do_and_retry(self, *args, **kwargs):
        """
        This method is a wrapper around Woob.do(), and handle interactive
        errors which allow to retry.

        List of handled errors:
        - BrowserQuestion
        - BrowserRedirect
        - DecoupledValidation
        """

        if self.stdout.isatty():
            # Set a non-None value to all backends's request_information
            #
            # - None indicates non-interactive: do not trigger 2FA challenges,
            #   raise NeedInteractive* exceptions before doing so
            # - non-None indicates interactive: ok to trigger 2FA challenges,
            #   raise BrowserQuestion/AppValidation when facing one
            # It should be a dict because when non-empty, it will contain HTTP
            # headers for legal PSD2 AIS/PIS authentication.
            for backend in self.enabled_backends:
                key = 'request_information'
                if key in backend.config and backend.config[key].get() is None:
                    backend.config[key].set({})

        try:
            for obj in self.do(*args, **kwargs):
                yield obj
        except CallErrors as errors:
            # Errors which are not handled here and which will be re-raised.
            remaining_errors = []
            # Backends on which we will retry.
            backends = set()

            for backend, error, backtrace in errors.errors:
                if isinstance(error, BrowserQuestion):
                    for field in error.fields:
                        v = self.ask(field)
                        backend.config[field.id].set(v)
                elif isinstance(error, BrowserRedirect):
                    print('Open this URL in a browser:')
                    print(error.url)
                    print()
                    value = self.ask('Please enter the final URL')
                    backend.config['auth_uri'].set(value)
                elif isinstance(error, DecoupledValidation):
                    print(error.message)
                    # FIXME we should reset this value, in case another DecoupledValidation occurs
                    key = 'resume'
                    if key in backend.config:
                        backend.config[key].set(True)
                else:
                    # Not handled error.
                    remaining_errors.append((backend, error, backtrace))
                    continue

                backends.add(backend)

            if backends:
                # There is at least one backend on which we can retry, do it
                # only on this ones.
                kwargs['backends'] = backends
                try:
                    for obj in self._do_and_retry(*args, **kwargs):
                        yield obj
                except CallErrors as sub_errors:
                    # As we called _do_and_retry, these sub errors are not
                    # interactive ones, so we can add them to the remaining
                    # errors.
                    remaining_errors += sub_errors.errors

            errors.errors = remaining_errors
            if errors.errors:
                # If there are remaining errors, raise them.
                raise errors

    # -- command tools ------------
    def parse_command_args(self, line, nb, req_n=None):
        try:
            args = shlex.split(line)
        except ValueError as e:
            raise ArgSyntaxError(str(e))

        if nb < len(args):
            raise TooManyArguments('Command takes at most %d arguments' % nb)
        if req_n is not None and (len(args) < req_n):
            raise NotEnoughArguments('Command needs %d arguments' % req_n)

        if len(args) < nb:
            args += tuple(None for i in range(nb - len(args)))
        return args

    # -- cmd.Cmd methods ---------
    def postcmd(self, stop, line):
        """
        This REPL method is overridden to return None instead of integers
        to prevent stopping cmdloop().
        """
        if not isinstance(stop, bool):
            stop = None
        return stop

    def parseline(self, line):
        """
        This REPL method is overridden to search "short" alias of commands
        """
        cmd, arg, ignored = super().parseline(line)

        if cmd is not None:
            names = set(name for name in self.get_names() if name.startswith('do_'))

            if 'do_' + cmd not in names:
                long = set(name for name in names if name.startswith('do_' + cmd))
                # if more than one result, ambiguous command, do nothing (error will display suggestions)
                if len(long) == 1:
                    cmd = long.pop()[3:]

        return cmd, arg, ignored

    def onecmd(self, line):
        """
        This REPL method is overridden to catch some particular exceptions.
        """
        cmd, arg, ignored = self.parseline(line)

        # Set the right formatter for the command.
        try:
            formatter_name = self.commands_formatters[cmd]
        except KeyError:
            formatter_name = self.DEFAULT_FORMATTER
        self.set_formatter(formatter_name)

        try:
            try:
                return super().onecmd(line)
            except CallErrors as e:
                return self.bcall_errors_handler(e)
            except BackendNotGiven as e:
                print('Error: %s' % str(e), file=self.stderr)
                return os.EX_DATAERR
            except NotEnoughArguments as e:
                print('Error: not enough arguments. %s' % str(e), file=self.stderr)
                return os.EX_USAGE
            except TooManyArguments as e:
                print('Error: too many arguments. %s' % str(e), file=self.stderr)
                return os.EX_USAGE
            except ArgSyntaxError as e:
                print('Error: invalid arguments. %s' % str(e), file=self.stderr)
                return os.EX_USAGE
            except (KeyboardInterrupt, EOFError):
                # ^C during a command process doesn't exit application.
                print('\nAborted.')
                return signal.SIGINT + 128
        finally:
            self.flush()

    def emptyline(self):
        """
        By default, an emptyline repeats the previous command.
        Overriding this function disables this behaviour.
        """
        pass

    def default(self, line):
        print('Unknown command: "%s"' % line, file=self.stderr)
        cmd, arg, ignore = super().parseline(line)
        if cmd is not None:
            names = set(name[3:] for name in self.get_names() if name.startswith('do_' + cmd))
            if len(names) > 0:
                print('Do you mean: %s?' % ', '.join(names), file=self.stderr)
        return os.EX_USAGE

    def completenames(self, text, *ignored):
        return [name for name in super().completenames(text, *ignored) if name not in self.hidden_commands]

    def _shell_completion_items(self):
        items = super()._shell_completion_items()
        items.update(
            set(self.completenames('')) -
            set(('debug', 'condition', 'count', 'formatter', 'logging', 'select', 'quit')))
        return items

    def path_completer(self, arg):
        dirname = os.path.dirname(arg)
        try:
            children = os.listdir(dirname or '.')
        except OSError:
            return ()
        l = []
        for child in children:
            path = os.path.join(dirname, child)
            if os.path.isdir(path):
                child += '/'
            l.append(child)
        return l

    def complete(self, text, state):
        """
        Override of the Cmd.complete() method to:

          * add a space at end of proposals
          * display only proposals for words which match the
            text already written by user.
        """
        super().complete(text, state)

        # When state = 0, Cmd.complete() set the 'completion_matches' attribute by
        # calling the completion function. Then, for other states, it only tries to
        # get the right item in list.
        # So that's the good place to rework the choices.
        if state == 0:
            self.completion_matches = [choice for choice in self.completion_matches if choice.startswith(text)]

        try:
            match = self.completion_matches[state]
        except IndexError:
            return None
        else:
            if match[-1] != '/':
                return '%s ' % match
            return match

    # -- errors management -------------
    def bcall_error_handler(self, backend, error, backtrace):
        """
        Handler for an exception inside the CallErrors exception.

        This method can be overridden to support more exceptions types.
        """
        return super().bcall_error_handler(backend, error, backtrace)

    def bcall_errors_handler(self, errors, ignore=()):
        if self.interactive:
            return super().bcall_errors_handler(errors, 'Use "logging debug" option to print backtraces.', ignore)
        else:
            return super().bcall_errors_handler(errors, ignore=ignore)

    # -- options related methods -------------
    def _handle_options(self):
        if self.options.formatter:
            self.commands_formatters = {}
            self.DEFAULT_FORMATTER = self.options.formatter
        self.set_formatter(self.DEFAULT_FORMATTER)

        if self.options.select:
            self.selected_fields = self.options.select.split(',')
        else:
            self.selected_fields = ['$direct']


        if self.options.count is not None:
            self._is_default_count = False
            if self.options.count <= 0:
                # infinite search
                self.options.count = None

        if self.options.condition:
            self.condition = ResultsCondition(self.options.condition)
        else:
            self.condition = None

        return super()._handle_options()

    def get_command_help(self, command, short=False):
        try:
            func = getattr(self, 'do_' + command)
        except AttributeError:
            return None

        doc = func.__doc__
        assert doc is not None, "A command must have a docstring"

        lines = [line.strip() for line in doc.strip().split('\n')]
        if not lines[0].startswith(command):
            lines = [command, ''] + lines

        if short:
            return lines[0]

        return '\n'.join(lines)

    def get_commands_doc(self):
        names = set(name for name in self.get_names() if name.startswith('do_'))
        appname = self.APPNAME.capitalize()
        d = OrderedDict(((appname, []), ('Woob', [])))

        for name in sorted(names):
            cmd = name[3:]
            if cmd in self.hidden_commands.union(self.woob_commands).union(['help']):
                continue

            d[appname].append(self.get_command_help(cmd))
        if not self.DISABLE_REPL:
            for cmd in self.woob_commands:
                d['Woob'].append(self.get_command_help(cmd))

        return d

    # -- default REPL commands ---------
    def do_quit(self, arg):
        """
        Quit the application.
        """
        return True

    def do_EOF(self, arg):
        """
        Quit the command line interpreter when ^D is pressed.
        """
        # print empty line for the next shell prompt to appear on the first column of the terminal
        print()
        return self.do_quit(arg)

    def do_help(self, arg=None):
        """
        help [COMMAND]

        List commands, or get information about a command.
        """
        if arg:
            cmd_names = set(name[3:] for name in self.get_names() if name.startswith('do_'))
            if arg in cmd_names:
                command_help = self.get_command_help(arg)
                if command_help is None:
                    logging.warning('Command "%s" is undocumented' % arg)
                else:
                    lines = command_help.split('\n')
                    lines[0] = '%s%s%s' % (self.BOLD, lines[0], self.NC)
                    self.stdout.write('%s\n' % '\n'.join(lines))
            else:
                print('Unknown command: "%s"' % arg, file=self.stderr)
        else:
            cmds = self._parser.formatter.format_commands(self._parser.commands)
            self.stdout.write('%s\n' % cmds)
            self.stdout.write('Type "help <command>" for more info about a command.\n')
        return 2

    def complete_backends(self, text, line, begidx, endidx):
        choices = []
        commands = ['enable', 'disable', 'only', 'list', 'add', 'register', 'edit', 'remove', 'list-modules']
        available_backends_names = set(backend.name for backend in self.woob.iter_backends())
        enabled_backends_names = set(backend.name for backend in self.enabled_backends)

        args = line.split(' ')
        if len(args) == 2:
            choices = commands
        elif len(args) >= 3:
            if args[1] == 'enable':
                choices = sorted(available_backends_names - enabled_backends_names)
            elif args[1] == 'only':
                choices = sorted(available_backends_names)
            elif args[1] == 'disable':
                choices = sorted(enabled_backends_names)
            elif args[1] in ('add', 'register') and len(args) == 3:
                for name, module in sorted(self.woob.repositories.get_all_modules_info(self.CAPS).items()):
                    choices.append(name)
            elif args[1] == 'edit':
                choices = sorted(available_backends_names)
            elif args[1] == 'remove':
                choices = sorted(available_backends_names)

        return choices

    def do_storage(self, line: str) -> int:
        """
        storage [ACTION] [BACKEND_NAME]

        Manipulation of a backend storage.

            * cat           display content of the storage
            * edit          edit the storage
            * flush         flush storage
        """
        args = line.split()
        if len(args) != 2:
            self.do_help('storage')
            return 1

        try:
            backend = self.woob.get_backend(args[1])
        except KeyError:
            print('No such backend: {args[1]}', file=self.stderr)
            return 1

        if args[0] == 'cat':
            storage = backend.storage.get()
            print(yaml.dump(storage), end='')

            return 0

        if args[0] == 'edit':
            storage = backend.storage.get()
            output = self.acquire_input(yaml.dump(storage), suffix='.yml')

            storage = yaml.safe_load(output)

            backend.storage.set(storage)
            backend.storage.save()

            print(f'Storage for backend {backend.name} has been saved.')
            return 0

        if args[0] == 'flush':
            backend.storage.set({})
            backend.storage.save()

            print(f'Storage for backend {backend.name} has been flushed.')
            return 0

        print(f'Action not found: {args[0]}', file=self.stderr)
        return 1

    def do_backends(self, line):
        """
        backends [ACTION] [BACKEND_NAME]...

        Select used backends.

        ACTION is one of the following (default: list):
            * enable         enable given backends
            * disable        disable given backends
            * only           enable given backends and disable the others
            * list           list backends
            * add            add a backend
            * register       register a new account on a website
            * edit           edit a backend
            * remove         remove a backend
            * list-modules   list modules
        """
        line = line.strip()
        if line:
            args = line.split()
        else:
            args = ['list']

        action = args[0]
        given_backend_names = args[1:]

        for backend_name in given_backend_names:
            if action in ('add', 'register'):
                minfo = self.woob.repositories.get_module_info(backend_name)
                if minfo is None:
                    print('Module "%s" does not exist.' % backend_name, file=self.stderr)
                    return 1
                else:
                    if not minfo.has_caps(self.CAPS):
                        print('Module "%s" is not supported by this application => skipping.' % backend_name, file=self.stderr)
                        return 1
            else:
                if backend_name not in [backend.name for backend in self.woob.iter_backends()]:
                    print('Backend "%s" does not exist => skipping.' % backend_name, file=self.stderr)
                    return 1

        if action in ('enable', 'disable', 'only', 'add', 'register', 'edit', 'remove'):
            if not given_backend_names:
                print('Please give at least a backend name.', file=self.stderr)
                return 2

        given_backends = set(backend for backend in self.woob.iter_backends() if backend.name in given_backend_names)

        if action == 'enable':
            for backend in given_backends:
                self.enabled_backends.add(backend)
        elif action == 'disable':
            for backend in given_backends:
                try:
                    self.enabled_backends.remove(backend)
                except KeyError:
                    print('%s is not enabled' % backend.name, file=self.stderr)
        elif action == 'only':
            self.enabled_backends = set()
            for backend in given_backends:
                self.enabled_backends.add(backend)
        elif action == 'list':
            enabled_backends_names = set(backend.name for backend in self.enabled_backends)
            disabled_backends_names = set(backend.name for backend in self.woob.iter_backends()) - enabled_backends_names
            print('Enabled: %s' % ', '.join(enabled_backends_names))
            if len(disabled_backends_names) > 0:
                print('Disabled: %s' % ', '.join(disabled_backends_names))
        elif action == 'add':
            for name in given_backend_names:
                instname = self.add_backend(name, name)
                if instname:
                    self.load_backends(names=[instname])
        elif action == 'register':
            for name in given_backend_names:
                instname = self.register_backend(name)
                if isinstance(instname, str):
                    self.load_backends(names=[instname])
        elif action == 'edit':
            for backend in given_backends:
                enabled = backend in self.enabled_backends
                self.unload_backends(names=[backend.name])
                self.edit_backend(backend.name)
                for newb in self.load_backends(names=[backend.name]).values():
                    if not enabled:
                        self.enabled_backends.remove(newb)
        elif action == 'remove':
            for backend in given_backends:
                self.woob.backends_config.remove_backend(backend.name)
                self.unload_backends(backend.name)
        elif action == 'list-modules':
            modules = []
            print('Modules list:')
            for name, info in sorted(self.woob.repositories.get_all_modules_info().items()):
                if not self.is_module_loadable(info):
                    continue
                modules.append(name)
                loaded = ' '
                for bi in self.woob.iter_backends():
                    if bi.NAME == name:
                        if loaded == ' ':
                            loaded = 'X'
                        elif loaded == 'X':
                            loaded = 2
                        else:
                            loaded += 1
                print(
                    '[%s] %s%s' % (
                        colored(str(loaded), attrs=['bold']),
                        colored('%-20s' % name, 'magenta', attrs=['bold']),
                        colored(info.description, 'green')
                    )
                )

        else:
            print('Unknown action: "%s"' % action, file=self.stderr)
            return 1

        if len(self.enabled_backends) == 0:
            print('Warning: no more backends are loaded. %s is probably unusable.' % self.APPNAME.capitalize(), file=self.stderr)

    def complete_logging(self, text, line, begidx, endidx):
        levels = ('debug', 'info', 'warning', 'error', 'quiet', 'default')
        args = line.split(' ')
        if len(args) == 2:
            return levels
        return ()

    def do_logging(self, line):
        """
        logging [LEVEL]

        Set logging level.

        Availables: debug, info, warning, error.
        * quiet is an alias for error
        * default is an alias for warning
        """
        args = self.parse_command_args(line, 1, 0)
        levels = (('debug',   logging.DEBUG),
                  ('info',    logging.INFO),
                  ('warning', logging.WARNING),
                  ('error',   logging.ERROR),
                  ('quiet',   logging.ERROR),
                  ('default', logging.WARNING)
                 )

        if not args[0]:
            current = None
            for label, level in levels:
                if logging.root.level == level:
                    current = label
                    break
            print('Current level: %s' % current)
            return

        levels = dict(levels)
        try:
            level = levels[args[0]]
        except KeyError:
            print('Level "%s" does not exist.' % args[0], file=self.stderr)
            print('Availables: %s' % ' '.join(levels), file=self.stderr)
            return 2
        else:
            logging.root.setLevel(level)
            for handler in logging.root.handlers:
                handler.setLevel(level)

    def do_condition(self, line):
        """
        condition [EXPRESSION | off]

        If an argument is given, set the condition expression used to filter the results. See CONDITION section for more details and the expression.
        If the "off" value is given, conditional filtering is disabled.

        If no argument is given, print the current condition expression.
        """
        line = line.strip()
        if line:
            if line == 'off':
                self.condition = None
            else:
                try:
                    self.condition = ResultsCondition(line)
                except ResultsConditionError as e:
                    print('%s' % e, file=self.stderr)
                    return 2
        else:
            if self.condition is None:
                print('No condition is set.')
            else:
                print(str(self.condition))

    def do_count(self, line):
        """
        count [NUMBER | off]

        If an argument is given, set the maximum number of results fetched.
        NUMBER must be at least 1.
        "off" value disables counting, and allows infinite searches.

        If no argument is given, print the current count value.
        """
        line = line.strip()
        if line:
            if line == 'off':
                self.options.count = None
                self._is_default_count = False
            else:
                try:
                    count = int(line)
                except ValueError:
                    print('Could not interpret "%s" as a number.' % line, file=self.stderr)
                    return 2
                else:
                    if count > 0:
                        self.options.count = count
                        self._is_default_count = False
                    else:
                        self.options.count = None
                        self._is_default_count = False
        else:
            if self.options.count is None:
                print('Counting disabled.')
            else:
                print(self.options.count)

    def complete_formatter(self, text, line, *ignored):
        formatters = self.formatters_loader.get_available_formatters()
        commands = ['list', 'option'] + formatters
        options = ['header', 'keys']
        option_values = ['on', 'off']

        args = line.split(' ')
        if len(args) == 2:
            return commands
        if args[1] == 'option':
            if len(args) == 3:
                return options
            if len(args) == 4:
                return option_values
        elif args[1] in formatters:
            return list(set(name[3:] for name in self.get_names() if name.startswith('do_')))

    def do_formatter(self, line):
        """
        formatter [list | FORMATTER [COMMAND] | option OPTION_NAME [on | off]]

        If a FORMATTER is given, set the formatter to use.
        You can add a COMMAND to apply the formatter change only to
        a given command.

        If the argument is "list", print the available formatters.

        If the argument is "option", set the formatter options.
        Valid options are: header, keys.
        If on/off value is given, set the value of the option.
        If not, print the current value for the option.

        If no argument is given, print the current formatter.
        """
        args = line.strip().split()
        if args:
            if args[0] == 'list':
                print(', '.join(self.formatters_loader.get_available_formatters()))
            elif args[0] == 'option':
                if len(args) > 1:
                    if len(args) == 2:
                        if args[1] == 'header':
                            print('off' if self.options.no_header else 'on')
                        elif args[1] == 'keys':
                            print('off' if self.options.no_keys else 'on')
                    else:
                        if args[2] not in ('on', 'off'):
                            print('Invalid value "%s". Please use "on" or "off" values.' % args[2], file=self.stderr)
                            return 2
                        else:
                            if args[1] == 'header':
                                self.options.no_header = True if args[2] == 'off' else False
                            elif args[1] == 'keys':
                                self.options.no_keys = True if args[2] == 'off' else False
                else:
                    print('Don\'t know which option to set. Available options: header, keys.', file=self.stderr)
                    return 2
            else:
                if args[0] in self.formatters_loader.get_available_formatters():
                    if len(args) > 1:
                        self.commands_formatters[args[1]] = self.set_formatter(args[0])
                    else:
                        self.commands_formatters = {}
                        self.DEFAULT_FORMATTER = self.set_formatter(args[0])
                else:
                    print('Formatter "%s" is not available.\n'
                          'Available formatters: %s.' % (args[0], ', '.join(self.formatters_loader.get_available_formatters())), file=self.stderr)
                    return 1
        else:
            print('Default formatter: %s' % self.DEFAULT_FORMATTER)
            for key, klass in self.commands_formatters.items():
                print('Command "%s": %s' % (key, klass))

    def do_select(self, line):
        """
        select [FIELD_NAME]... | "$direct" | "$full"

        If an argument is given, set the selected fields.
        $direct selects all fields loaded in one http request.
        $full selects all fields using as much http requests as necessary.

        If no argument is given, print the currently selected fields.
        """
        line = line.strip()
        if line:
            split = line.split()
            self.selected_fields = split
        else:
            print(' '.join(self.selected_fields))

    # First sort in alphabetical of backend
    # Second, sort with ID
    def comp_key(self, obj):
        return (obj.backend, obj.id)

    @defaultcount(40)
    def do_ls(self, line):
        """
        ls [-d] [-U] [PATH]

        List objects in current path.
        If an argument is given, list the specified path.
        Use -U option to not sort results. It allows you to use a "fast path" to
        return results as soon as possible.
        Use -d option to display information about a collection (and to not
        display the content of it). It has the same behavior than the well
        known UNIX "ls" command.
        """
        # TODO: real parsing of options
        path = line.strip()
        only = False
        sort = True

        if '-U' in line.strip().partition(' '):
            path = line.strip().partition(' ')[-1]
            sort = False

        if '-d' in line.strip().partition(' '):
            path = None
            only = line.strip().partition(' ')[-1]

        if path:
            for _path in path.split('/'):
                # We have an argument, let's ch to the directory before the ls
                self.working_path.cd1(_path)

        objects = []
        collections = []
        self.objects = []

        self.start_format()

        for res in self._fetch_objects(objs=self.COLLECTION_OBJECTS):
            if isinstance(res, Collection):
                collections.append(res)
                if sort is False:
                    self.formatter.format_collection(res, only)
            else:
                if sort:
                    objects.append(res)
                else:
                    self._format_obj(res, only)

        if sort:
            objects.sort(key=self.comp_key)
            collections = self._merge_collections_with_same_path(collections)
            collections.sort(key=self.comp_key)
            for collection in collections:
                self.formatter.format_collection(collection, only)
            for obj in objects:
                self._format_obj(obj, only)

        if path:
            for _path in path.split('/'):
                # Let's go back to the parent directory
                self.working_path.up()
        else:
            # Save collections only if we listed the current path.
            self.collections = collections

    def _find_collection(self, collection, collections):
        for col in collections:
            if col.split_path == collection.split_path:
                return col
        return None

    def _merge_collections_with_same_path(self, collections):
        to_return = []
        for collection in collections:
            col = self._find_collection(collection, to_return)
            if col:
                col.backend += " %s" % collection.backend
            else:
                to_return.append(collection)
        return to_return

    def _format_obj(self, obj, only):
        if only is False or not hasattr(obj, 'id') or obj.id in only:
            self.cached_format(obj)


    def do_cd(self, line):
        """
        cd [PATH]

        Follow a path.
        ".." is a special case and goes up one directory.
        "" is a special case and goes home.
        """
        if not len(line.strip()):
            self.working_path.home()
        elif line.strip() == '..':
            self.working_path.up()
        else:
            self.working_path.cd1(line)

            collections = []
            try:
                for res in self.do('get_collection', objs=self.COLLECTION_OBJECTS,
                                   split_path=self.working_path.get(),
                                   caps=CapCollection):
                    if res:
                        collections.append(res)
            except CallErrors as errors:
                self.bcall_errors_handler(errors, CollectionNotFound)

            if len(collections):
                # update the path from the collection if possible
                if len(collections) == 1:
                    self.working_path.split_path = collections[0].split_path
            else:
                print("Path: %s not found" % self.working_path, file=self.stderr)
                self.working_path.restore()
                return 1

        self._change_prompt()

    def _fetch_objects(self, objs):
        split_path = self.working_path.get()

        try:
            for res in self._do_and_retry(
                'iter_resources',
                objs=objs,
                split_path=split_path,
                caps=CapCollection
            ):
                yield res
        except CallErrors as errors:
            self.bcall_errors_handler(errors, CollectionNotFound)


    def all_collections(self):
        """
        Get all objects that are collections: regular objects and fake dumb objects.
        """
        obj_collections = [obj for obj in self.objects if isinstance(obj, BaseCollection)]
        return obj_collections + self.collections

    def obj_to_filename(self, obj, dest=None, default=None):
        """
        This method can be used to get a filename from an object, using a mask
        filled by information of this object.

        All patterns are braces-enclosed, and are name of available fields in
        the object.

        :param obj: object
        :type obj: BaseObject
        :param dest: dest given by user (default None)
        :type dest: str
        :param default: default file mask (if not given, this is '{id}-{title}.{ext}')
        :type default: str
        :rtype: str
        """
        if default is None:
            default = '{id}-{title}.{ext}'
        if dest is None:
            dest = '.'
        if os.path.isdir(dest):
            dest = os.path.join(dest, default)

        def repl(m):
            field = m.group(1)
            if hasattr(obj, field):
                value = getattr(obj, field)
                if empty(value):
                    value = 'unknown'
                return re.sub('[?:/]', '-', '%s' % value)
            else:
                return m.group(0)
        return re.sub(r'\{(.+?)\}', repl, dest)

    # for cd & ls
    def complete_path(self, text, line, begidx, endidx):
        directories = set()
        if len(self.working_path.get()):
            directories.add('..')
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)

        # refresh only if needed
        if len(self.objects) == 0 and len(self.collections) == 0:
            try:
                self.objects, self.collections = self._fetch_objects(objs=self.COLLECTION_OBJECTS)
            except CallErrors as errors:
                self.bcall_errors_handler(errors, CollectionNotFound)

        collections = self.all_collections()
        for collection in collections:
            directories.add(collection.basename)

        return [s[offs:] for s in directories if s.startswith(mline)]

    def complete_ls(self, text, line, begidx, endidx):
        return self.complete_path(text, line, begidx, endidx)

    def complete_cd(self, text, line, begidx, endidx):
        return self.complete_path(text, line, begidx, endidx)

    # -- formatting related methods -------------
    def set_formatter(self, name):
        """
        Set the current formatter from name.

        It returns the name of the formatter which has been really set.
        """
        try:
            self.formatter = self.formatters_loader.build_formatter(name)
        except FormatterLoadError as e:
            print('%s' % e, file=self.stderr)
            if self.DEFAULT_FORMATTER == name:
                self.DEFAULT_FORMATTER = ReplApplication.DEFAULT_FORMATTER
            print('Falling back to "%s".' % (self.DEFAULT_FORMATTER), file=self.stderr)
            self.formatter = self.formatters_loader.build_formatter(self.DEFAULT_FORMATTER)
            name = self.DEFAULT_FORMATTER
        if self.options.no_header:
            self.formatter.display_header = False
        if self.options.no_keys:
            self.formatter.display_keys = False
        if self.options.outfile:
            self.formatter.outfile = self.options.outfile
        if self.interactive:
            self.formatter.interactive = True
        return name

    def set_formatter_header(self, string):
        pass

    def start_format(self, **kwargs):
        self.formatter.start_format(**kwargs)

    def cached_format(self, obj):
        self.add_object(obj)
        alias = None
        if self.interactive:
            alias = '%s' % len(self.objects)
        self.format(obj, alias=alias)

    def parse_fields(self, fields):
        if '$direct' in fields:
            return []
        if '$full' in fields:
            return None
        return fields

    def format(self, result, alias=None):
        fields = self.parse_fields(self.selected_fields)
        try:
            self.formatter.format(obj=result, selected_fields=fields, alias=alias)
        except FieldNotFound as e:
            print(e, file=self.stderr)
        except MandatoryFieldsNotFound as e:
            print('%s Hint: select missing fields or use another formatter (ex: multiline).' % e, file=self.stderr)

    def flush(self):
        self.formatter.flush()

    def do_debug(self, line):
        """
        debug

        Launch a debug Python shell
        """

        from woob.applications.debug import AppDebug

        app = AppDebug()
        locs = dict(application=self, woob=self.woob)
        if len(self.woob.backend_instances):
            locs['backend'] = next(iter(self.woob.backend_instances.values()))
            locs['browser'] = locs['backend'].browser

        banner = ('Woob debug shell\n\nAvailable variables:\n'
         + '\n'.join(['  %s: %s' % (k, v) for k, v in locs.items()]))

        funcs = [app.ipython, app.bpython, app.python]
        app.launch(funcs, locs, banner)



# Copyright(C) 2010-2013 Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .base import (
    BaseObject, StringField, IntField, FloatField, Field, EnumField,
    Enum,
)
from .collection import CapCollection, CollectionNotFound, Collection
from .date import DateField
from .address import compat_field, PostalAddress

from datetime import time, datetime
from woob.tools.date import parse_date

__all__ = ['BaseCalendarEvent', 'CapCalendarEvent']


class CATEGORIES(Enum):
    CONCERT = 'Concert'
    CINE = 'Cinema'
    THEATRE = 'Theatre'
    TELE = 'Television'
    CONF = 'Conference'
    AUTRE = 'Autre'
    EXPO = 'Exposition'
    SPECTACLE = 'Spectacle'
    FEST = 'Festival'
    SPORT = 'Sport'


#the following elements deal with ICalendar stantdards
#see https://fr.wikipedia.org/wiki/ICalendar#%C3%89v%C3%A9nements_(VEVENT)
class TRANSP(Enum):
    OPAQUE = 'OPAQUE'
    TRANSPARENT = 'TRANSPARENT'


class STATUS(Enum):
    TENTATIVE = 'TENTATIVE'
    CONFIRMED = 'CONFIRMED'
    CANCELLED = 'CANCELLED'


class TICKET(Enum):
    AVAILABLE = 'Available'
    NOTAVAILABLE = 'Not available'
    CLOSED = 'Closed'


class BaseCalendarEvent(BaseObject):
    """
    Represents a calendar event
    """

    start_date = DateField('Start date of the event')
    end_date = DateField('End date of the event')
    timezone = StringField('Timezone of the event in order to convert to utc time', default='Etc/UCT')
    summary = StringField('Title of the event')
    address = Field('Address where event will take place', PostalAddress)
    category = EnumField('Category of the event', CATEGORIES)
    description = StringField('Description of the event')
    price = FloatField('Price of the event')
    booked_entries = IntField('Entry number')
    max_entries = IntField('Max entry number')
    event_planner = StringField('Name of the event planner')

    city = compat_field('address', 'city')
    location = compat_field('address', 'street')

    #the following elements deal with ICalendar stantdards
    #see https://fr.wikipedia.org/wiki/ICalendar#%C3%89v%C3%A9nements_(VEVENT)
    sequence = IntField('Number of updates, the first is number 1')

    # (TENTATIVE, CONFIRMED, CANCELLED)
    status = EnumField('Status of the event', STATUS)
    # (OPAQUE, TRANSPARENT)
    transp = EnumField('Describes if event is available', TRANSP)
    # (AVAILABLE, NOTAVAILABLE, CLOSED)
    ticket = EnumField('Describes if tickets are available', TICKET)

    @classmethod
    def id2url(cls, _id):
        """Overloaded in child classes provided by backends."""
        raise NotImplementedError()

    @property
    def page_url(self):
        """
        Get page URL of the announce.
        """
        return self.id2url(self.id)


class Query(BaseObject):
    """
    Query to find events
    """

    start_date = DateField('Start date of the event')
    end_date = DateField('End date of the event')
    city = StringField('Name of the city in witch event will take place')
    categories = Field('List of categories of the event', list, tuple, default=list(CATEGORIES))
    ticket = Field('List of status of the tickets sale', list, tuple, default=list(TICKET))
    summary = StringField('Title of the event')


class CapCalendarEvent(CapCollection):
    """
    Capability of calendar event type sites
    """

    ASSOCIATED_CATEGORIES = 'ALL'

    def has_matching_categories(self, query):
        if self.ASSOCIATED_CATEGORIES == 'ALL':
            return True

        for category in query.categories:
            if category in self.ASSOCIATED_CATEGORIES:
                return True
        return False

    def search_events(self, query):
        """
        Search event

        :param query: search query
        :type query: :class:`Query`
        :rtype: iter[:class:`BaseCalendarEvent`]
        """
        raise NotImplementedError()

    def list_events(self, date_from, date_to=None):
        """
        list coming event.

        :param date_from: date of beguinning of the events list
        :type date_from: date
        :param date_to: date of ending of the events list
        :type date_to: date
        :rtype: iter[:class:`BaseCalendarEvent`]
        """
        raise NotImplementedError()

    def get_event(self, _id):
        """
        Get an event from an ID.

        :param _id: id of the event
        :type _id: str
        :rtype: :class:`BaseCalendarEvent` or None is fot found.
        """
        raise NotImplementedError()

    def attends_event(self, event, is_attending=True):
        """
        Attends or not to an event
        :param event : the event
        :type event : BaseCalendarEvent
        :param is_attending : is attending to the event or not
        :type is_attending : bool
        """
        raise NotImplementedError()

    def iter_resources(self, objs, split_path):
        """
        Iter events by category
        """
        if len(split_path) == 0 and self.ASSOCIATED_CATEGORIES != 'ALL':
            for category in self.ASSOCIATED_CATEGORIES:
                collection = Collection([category], category)
                yield collection

        elif len(split_path) == 1 and split_path[0] in self.ASSOCIATED_CATEGORIES:
            query = Query()
            query.categories = split_path
            query.start_date = datetime.combine(parse_date('today'), time.min)
            query.end_date = parse_date('')
            query.city = ''
            for event in self.search_events(query):
                yield event

    def validate_collection(self, objs, collection):
        """
        Validate Collection
        """
        if collection.path_level == 0:
            return
        if collection.path_level == 1 and collection.split_path[0] in CATEGORIES.values:
            return
        raise CollectionNotFound(collection.split_path)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime, date

from .base import Capability, BaseObject, Field, FloatField, \
                  StringField, IntField, UserError, NotLoaded, EnumField, Enum
from .date import DateField

__all__ = [
    'Forecast', 'Current', 'City', 'CityNotFound', 'Temperature', 'CapWeather',
    'BaseWeather', 'Direction', 'Precipitation',
]


class Direction(Enum):
    S = 'South'
    N = 'North'
    E = 'East'
    W = 'West'
    SE = 'Southeast'
    SW = 'Southwest'
    NW = 'Northwest'
    NE = 'Northeast'
    SSE = 'South-Southeast'
    SSW = 'South-Southwest'
    NNW = 'North-Northwest'
    NNE = 'North-Northeast'
    ESE = 'East-Southeast'
    ENE = 'East-Northeast'
    WSW = 'West-Southwest'
    WNW = 'West-Northwest'
    Variable = 'Variable'


# METAR keys
class Precipitation(Enum):
    RA = 'Rain'
    SN = 'Snow'
    GR = 'Hail'
    PL = 'Ice pellets'
    GS = 'Small hail'
    DZ = 'Drizzle'
    IC = 'Ice cristals'
    SG = 'Small grains'
    UP = 'Unknown precipiation'


class Temperature(BaseObject):
    value =      FloatField('Temperature value')
    unit =       StringField('Input unit')

    def __init__(self, value=NotLoaded, unit = '', url=None):
        super(Temperature, self).__init__(str(value), url)
        self.value = value
        if unit not in ['C', 'F']:
            unit = ''
        self.unit = unit

    def asfahrenheit(self):
        if not self.unit:
            return '%s' % int(round(self.value))
        elif self.unit == 'F':
            return '%s °F' % int(round(self.value))
        else:
            return '%s °F' % int(round((self.value * 9.0 / 5.0) + 32))

    def ascelsius(self):
        if not self.unit:
            return '%s' % int(round(self.value))
        elif self.unit == 'C':
            return '%s °C' % int(round(self.value))
        else:
            return '%s °C' % int(round((self.value - 32.0) * 5.0 / 9.0))

    def __repr__(self):
        if self.value is not None and self.unit:
            return '%r %r' % (self.value, self.unit)
        return ''


class BaseWeather(BaseObject):
    precipitation = EnumField('Precipitation type', Precipitation)
    precipitation_probability = FloatField('Probability of precipitation (ratio)')

    wind_direction = EnumField('Wind direction', Direction)
    wind_speed = FloatField('Wind speed (in km/h)')

    humidity = FloatField('Relative humidity (ratio)')
    pressure = FloatField('Atmospheric pressure (in hPa)')

    visibility = FloatField('Horizontal visibility distance (in km)')
    cloud = IntField('Cloud coverage (in okta (0-8))')


class Forecast(BaseWeather):
    """
    Weather forecast.
    """
    date =      Field('Date for the forecast', datetime, date, str)
    low =       Field('Low temperature', Temperature)
    high =      Field('High temperature', Temperature)
    text =      StringField('Comment on forecast')

    def __init__(self, date=NotLoaded, low=None, high=None, text=None, unit=None, url=None):
        super(Forecast, self).__init__(str(date or ""), url)
        self.date = date
        self.low = Temperature(low, unit)
        self.high = Temperature(high, unit)
        self.text = text


class Current(BaseWeather):
    """
    Current weather.
    """
    date =      DateField('Date of measure')
    text =      StringField('Comment about current weather')
    temp =      Field('Current temperature', Temperature)

    def __init__(self, date=NotLoaded, temp=None, text=None, unit=None, url=None):
        super(Current, self).__init__(str(date or ""), url)
        self.date = date
        self.text = text
        self.temp = Temperature(temp, unit)


class City(BaseObject):
    """
    City where to find weather.
    """
    name =      StringField('Name of city')

    def __init__(self, id='', name=None, url=None):
        super(City, self).__init__(id, url)
        self.name = name


class CityNotFound(UserError):
    """
    Raised when a city is not found.
    """


class CapWeather(Capability):
    """
    Capability for weather websites.
    """

    def iter_city_search(self, pattern):
        """
        Look for a city.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`City`]
        """
        raise NotImplementedError()

    def get_current(self, city_id):
        """
        Get current weather.

        :param city_id: ID of the city
        :rtype: :class:`Current`
        """
        raise NotImplementedError()

    def iter_forecast(self, city_id):
        """
        Iter forecasts of a city.

        :param city_id: ID of the city
        :rtype: iter[:class:`Forecast`]
        """
        raise NotImplementedError()



# Copyright(C) 2016-2021 Edouard Lefebvre du Prey
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import dbm.ndbm
import yaml

from .iconfig import ConfigError, IConfig
from .yamlconfig import WoobDumper




__all__ = ['DBMConfig']


class DBMConfig(IConfig):
    def __init__(self, path):
        self.path = path
        self.storage = None

    def load(self, default=None):
        self.storage = dbm.ndbm.open(self.path, 'c')

    def save(self):
        if hasattr(self.storage, 'sync'):
            self.storage.sync()

    def get(self, *args, **kwargs):
        key = '.'.join(args)
        try:
            value = self.storage[key]
            value = yaml.load(value, Loader=yaml.SafeLoader)
        except KeyError as exc:
            if 'default' in kwargs:
                value = kwargs.get('default')
            else:
                raise ConfigError() from exc
        except TypeError as exc:
            raise ConfigError() from exc
        return value

    def set(self, *args):
        key = '.'.join(args[:-1])
        value = args[-1]
        try:
            self.storage[key] = yaml.dump(value, None, Dumper=WoobDumper, default_flow_style=False)
        except KeyError as exc:
            raise ConfigError() from exc
        except TypeError as exc:
            raise ConfigError() from exc

    def delete(self, *args):
        key = '.'.join(args)
        try:
            del self.storage[key]
        except KeyError as exc:
            raise ConfigError() from exc
        except TypeError as exc:
            raise ConfigError() from exc



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, StringField, Field
from .address import GeoCoordinates, PostalAddress, compat_field


__all__ = ['IpLocation', 'CapGeolocIp']


class IpLocation(BaseObject):
    """
    Represents the location of an IP address.
    """

    address = Field('Address', PostalAddress)
    geo = Field('Geolocation', GeoCoordinates)

    osmlink =   StringField('Link to OpenStreetMap location page')
    host =      StringField('Hostname')
    tld =       StringField('Top Level Domain')
    isp =       StringField('Internet Service Provider')

    lt = compat_field('geo', 'latitude')
    lg = compat_field('geo', 'longitude')

    city = compat_field('address', 'city')
    region = compat_field('address', 'region')
    zipcode = compat_field('address', 'postal_code')
    country = compat_field('address', 'country')


class CapGeolocIp(Capability):
    """
    Access information about IP addresses database.
    """

    def get_location(self, ipaddr):
        """
        Get location of an IP address.

        :param ipaddr: IP address
        :type ipaddr: str
        :rtype: :class:`IpLocation`
        """
        raise NotImplementedError()



# Copyright(C) 2018 Phyks
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import itertools
from collections import Counter

from woob.capabilities.base import empty
from woob.capabilities.housing import POSTS_TYPES


class HousingTest:
    """
    Testing class to standardize the housing modules tests.
    """
    # Fields to be checked for values across all items in housings list
    FIELDS_ALL_HOUSINGS_LIST = [
        "id", "type", "advert_type", "house_type", "url", "title", "area",
        "cost", "currency", "utilities", "date", "location", "station", "text",
        "phone", "rooms", "bedrooms", "DPE", "GES", "details"
    ]
    # Fields to be checked for at least one item in housings list
    FIELDS_ANY_HOUSINGS_LIST = [
        "photos"
    ]
    # Fields to be checked for values across all items when querying
    # individually
    FIELDS_ALL_SINGLE_HOUSING = [
        "id", "url", "type", "advert_type", "house_type", "title", "area",
        "cost", "currency", "utilities", "date", "location", "station", "text",
        "phone", "rooms", "bedrooms", "DPE", "GES", "details"
    ]
    # Fields to be checked for values at least once for all items when querying
    # individually
    FIELDS_ANY_SINGLE_HOUSING = [
        "photos"
    ]
    # Some backends cannot distinguish between rent and furnished rent for
    # single housing post. Set this to True if this is the case.
    DO_NOT_DISTINGUISH_FURNISHED_RENT = False

    def assertNotEmpty(self, obj, field):
        self.assertFalse(
            empty(getattr(obj, field)),
            'Field "%s" is empty and should not be.' % field
        )

    def check_housing_lists(self, query, max_results=20):
        results = list(itertools.islice(
            self.backend.search_housings(query),
            max_results
        ))
        self.assertGreater(len(results), 0)

        for field in self.FIELDS_ANY_HOUSINGS_LIST:
            self.assertTrue(
                any(not empty(getattr(x, field)) for x in results),
                'Missing a "%s" field.' % field
            )

        for x in results:
            if 'type' in self.FIELDS_ALL_HOUSINGS_LIST:
                if (
                    self.DO_NOT_DISTINGUISH_FURNISHED_RENT and
                    query.type in [POSTS_TYPES.RENT, POSTS_TYPES.FURNISHED_RENT]
                ):
                    self.assertIn(x.type,
                                [POSTS_TYPES.RENT, POSTS_TYPES.FURNISHED_RENT])
                else:
                    self.assertEqual(x.type, query.type)
            if 'advert_type' in self.FIELDS_ALL_HOUSINGS_LIST:
                self.assertIn(x.advert_type, query.advert_types)
            if 'house_type' in self.FIELDS_ALL_HOUSINGS_LIST:
                self.assertIn(x.house_type, query.house_types)
            for field in self.FIELDS_ALL_HOUSINGS_LIST:
                self.assertNotEmpty(x, field)
            if not empty(x.cost):
                self.assertNotEmpty(x, 'price_per_meter')
            for photo in x.photos:
                self.assertRegex(photo.url, r'^http(s?)://')

        return results

    def check_single_housing_all(self, housing,
                                 type, house_type, advert_type):
        for field in self.FIELDS_ALL_SINGLE_HOUSING:
            self.assertNotEmpty(housing, field)
        if 'type' in self.FIELDS_ALL_SINGLE_HOUSING:
            if (
                self.DO_NOT_DISTINGUISH_FURNISHED_RENT and
                type in [POSTS_TYPES.RENT, POSTS_TYPES.FURNISHED_RENT]
            ):
                self.assertIn(housing.type,
                              [POSTS_TYPES.RENT, POSTS_TYPES.FURNISHED_RENT])
            else:
                self.assertEqual(housing.type, type)
        if 'house_type' in self.FIELDS_ALL_SINGLE_HOUSING:
            if not empty(house_type):
                self.assertEqual(housing.house_type, house_type)
            else:
                self.assertNotEmpty(housing, 'house_type')
        if 'advert_type' in self.FIELDS_ALL_SINGLE_HOUSING:
            self.assertEqual(housing.advert_type, advert_type)

    def check_single_housing_any(self, housing, counter):
        for field in self.FIELDS_ANY_SINGLE_HOUSING:
            if not empty(getattr(housing, field)):
                counter[field] += 1
        for photo in housing.photos:
            self.assertRegex(photo.url, r'^http(s?)://')
        return counter

    def check_against_query(self, query, max_results=20):
        # Check housing listing results
        results = self.check_housing_lists(query, max_results)

        # Check mandatory fields in all housings

        housing = self.backend.fillobj(results[0], self.FIELDS_ANY_SINGLE_HOUSING + self.FIELDS_ALL_SINGLE_HOUSING)

        self.check_single_housing_all(
            housing,
            results[0].type,
            results[0].house_type,
            results[0].advert_type
        )

        # Check fields that should appear in at least one housing
        counter = Counter()
        counter = self.check_single_housing_any(housing, counter)
        for result in results[1:]:
            if all(counter[field] > 0 for field in
                   self.FIELDS_ANY_SINGLE_HOUSING):
                break
            housing = self.backend.fillobj(result, self.FIELDS_ANY_SINGLE_HOUSING + self.FIELDS_ALL_SINGLE_HOUSING)

            counter = self.check_single_housing_any(housing, counter)
        for field in self.FIELDS_ANY_SINGLE_HOUSING:
            self.assertGreater(
                counter[field],
                0,
                'Optional field "%s" should appear at least once.' % field
            )



# Copyright(C) 2010-2021 Nicolas Duhamel
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from copy import copy
from posixpath import sep, join


class WorkingPath:
    def __init__(self):
        self.split_path = []
        self.previous = copy(self.split_path)

    def cd1(self, user_input):
        """
        Append *one* level to the current path.
        This means that separators (/) will get escaped.
        """
        split_path = self.get()
        split_path.append(user_input)
        self.location(split_path)

    def location(self, split_path):
        """
        Go to a new path, and store the previous path.
        """
        self.previous = self.get()
        self.split_path = split_path

    def restore(self):
        """
        Go to the previous path
        """
        self.split_path, self.previous = self.previous, self.split_path

    def home(self):
        """
        Go to the root
        """
        self.location([])

    def up(self):
        """
        Go up one directory
        """
        self.location(self.split_path[:-1])

    def get(self):
        """
        Get the current working path
        """
        return copy(self.split_path)

    def __str__(self):
        return join(sep, *[s.replace('/', r'\/') for s in self.split_path])
